import pkg_resources

from .exceptions import ArrException, ConnectionFailure, Excluded, Exists, Invalid, NotFound, Unauthorized
from .objs.simple import MetadataProfile, RemotePathMapping, RootFolder, UnmappedFolder, Season
from .objs.reload import QualityProfile, LanguageProfile, SystemStatus, Tag, Movie, Series
from .apis.sonarr import SonarrAPI
from .apis.radarr import RadarrAPI
from .apis.lidarr import LidarrAPI
from .apis.readarr import ReadarrAPI

try:
    __version__ = pkg_resources.get_distribution("arrapi").version
except pkg_resources.DistributionNotFound:
    __version__ = ""
__author__ = "Nathan Taggart"
__credits__ = "meisnate12"
__package_name__ = "arrapi"
__project_name__ = "ArrAPI"
__description__ = "Python wrapper for Radarr and Sonarr APIs:"
__url__ = "https://github.com/meisnate12/ArrAPI"
__email__ = 'meisnate12@gmail.com'
__license__ = 'MIT License'
__all__ = [
    "RadarrAPI",
    "SonarrAPI",
    "QualityProfile",
    "LanguageProfile",
    "MetadataProfile",
    "RemotePathMapping",
    "RootFolder",
    "UnmappedFolder",
    "SystemStatus",
    "Tag",
    "Movie",
    "Series",
    "Season",
    "ArrException",
    "ConnectionFailure",
    "Excluded",
    "Exists",
    "Invalid",
    "NotFound",
    "Unauthorized"
]



from requests import Session
from typing import Optional
from .base import BaseV1API
from ..raws.readarr import ReadarrRawAPI


class ReadarrAPI(BaseV1API):
    def __init__(self, url: str, apikey: str, session: Optional[Session] = None) -> None:
        super().__init__(ReadarrRawAPI(url, apikey, session=session))



class ArrException(Exception):
    """ Base class for all ArrAPI exceptions. """
    pass


class ConnectionFailure(ArrException):
    """ Failed to connect to Arr instance. """
    pass


class Excluded(ArrException):
    """ Item is excluded from being added to the Arr instance. """
    pass


class Exists(ArrException):
    """ Item already exists in the Arr instance. """
    pass


class Invalid(ArrException):
    """ Invalid Selection. """
    pass


class NotFound(ArrException):
    """ Item not found. """
    pass


class Unauthorized(ArrException):
    """ Invalid apikey. """
    pass



from requests import Session
from typing import Optional
from .base import BaseRawV1API

class LidarrRawAPI(BaseRawV1API):
    def __init__(self, url: str, apikey: str, session: Optional[Session] = None) -> None:
        super().__init__(url, apikey, session=session)



from abc import abstractmethod
from typing import Union, Optional, List, TYPE_CHECKING

from arrapi import NotFound, Invalid, Exists, Excluded
from arrapi.objs.base import BaseObj

if TYPE_CHECKING:
    from arrapi.objs.simple import RootFolder

class ReloadObj(BaseObj):
    def __init__(self, arr, data, load=False):
        super().__init__(arr, data)
        if load:
            self._load(None)

    @abstractmethod
    def _load(self, data):
        self._partial = data is not None
        super()._load(self._full_load() if data is None else data)

    @abstractmethod
    def _full_load(self):
        pass

    def reload(self):
        """ Reloads the Object """
        self._load(None)


class QualityProfile(ReloadObj):
    """ Represents a single Quality Profile.

        Attributes:
            id (int): ID of the Quality Profile.
            name (str): Name of the Quality Profile.
    """

    def _load(self, data):
        super()._load(data)
        self.id = self._parse(attrs="id", value_type="int")
        self.name = self._parse(attrs="name")
        self._finish(self.name)

    def _full_load(self):
        return self._raw.get_qualityProfileId(self.id)


class LanguageProfile(ReloadObj):
    """ Represents a single Language Profile.

        Attributes:
            id (int): ID of the Language Profile.
            name (str): Name of the Language Profile.
    """

    def _load(self, data):
        super()._load(data)
        self.id = self._parse(attrs="id", value_type="int")
        self.name = self._parse(attrs="name")
        self._finish(self.name)

    def _full_load(self):
        return self._raw.get_languageProfileId(self.id)

class SystemStatus(ReloadObj):
    """ Represents the System Status.

        Attributes:
            version (str): Version of the Arr Instance.

        Check dir(SystemStatus) for all attribute as the rest are auto built.
    """

    def __init__(self, arr):
        super().__init__(arr, None)

    def _load(self, data):
        super()._load(None)
        self.version = ""
        for key, value in self._data.items():
            if key.startswith("is") or key == "migrationVersion":
                setattr(self, key, self._parse(data=value, value_type="int"))
            elif key.endswith("Time"):
                setattr(self, key, self._parse(data=value, value_type="date"))
            elif key == "version":
                self.version = value
            else:
                setattr(self, key, value)
        self._finish(self.version)

    def _full_load(self):
        return self._raw.get_system_status()


class Tag(ReloadObj):
    """ Represents a single Tag.

        Attributes:
            id (int): ID of the Tag.
            label (str): Label of the Tag.
            detail (bool): If the Tag was loaded with details.
            delayProfileIds (List[int]): Delay Profile IDs. (Only when loaded with details)
            notificationIds (List[int]): Notification IDs. (Only when loaded with details)
            restrictionIds (List[int]): Restriction IDs. (Only when loaded with details)
            importListIds (List[int]): Import List IDs. (Only when loaded with details)
            movieIds (List[int]): Radarr Movie IDs. (Only when loaded with details using :class:`~arrapi.apis.radarr.RadarrAPI`)
            seriesIds (List[int]): Sonarr Series IDs. (Only when loaded with details using :class:`~arrapi.apis.radarr.SonarrAPI`)
            artistIds (List[int]): Lidarr Artist IDs. (Only when loaded with details using :class:`~arrapi.apis.lidarr.LidarrAPI`)
            authorIds (List[int]): Readarr Author IDs. (Only when loaded with details using :class:`~arrapi.apis.readarr.ReadarrAPI`)
    """

    def _load(self, data):
        super()._load(data)
        self.label = self._parse(attrs="label")
        self.id = self._parse(attrs="id", value_type="int", default_is_none=True)
        self.detail = False
        if "delayProfileIds" in self._data:
            self.detail = True
            self.delayProfileIds = self._parse(attrs="delayProfileIds", value_type="int", is_list=True)
            self.notificationIds = self._parse(attrs="notificationIds", value_type="int", is_list=True)
            self.restrictionIds = self._parse(attrs="restrictionIds", value_type="int", is_list=True)
            self.importListIds = self._parse(attrs="importListIds", value_type="int", is_list=True)
        if "movieIds" in self._data:
            self.movieIds = self._parse(attrs="movieIds", value_type="int", is_list=True)
        if "seriesIds" in self._data:
            self.seriesIds = self._parse(attrs="seriesIds", value_type="int", is_list=True)
        if "artistIds" in self._data:
            self.artistIds = self._parse(attrs="artistIds", value_type="int", is_list=True)
        if "authorIds" in self._data:
            self.authorIds = self._parse(attrs="authorIds", value_type="int", is_list=True)
        self._finish(self.label)

    def __str__(self):
        return f"[{self.id}:{self._name}]" if self.id and self._name else f"[Tag:{self.id}]"

    def _full_load(self):
        return self._raw.get_tag_id(self.id, detail=self.detail)

    def edit(self, label: str) -> None:
        """ Edit the :class:`~arrapi.objs.reload.Tag`.

            Parameters:
                label (str): Label to change tag to.

        """
        self._load(self._raw.put_tag_id(self.id, label))

    def delete(self) -> None:
        """ Delete the :class:`~arrapi.objs.reload.Tag`."""
        self._raw.delete_tag_id(self.id)


class Command(ReloadObj):
    """ Represents a single Command.

        Attributes:
            name (str): Name.
            commandName (str): Full named.
            message (str): Status Message.
            body (dict): Body of dict sent with the command.
            priority (str): Priority level.
            status (str): Status state.
            queued (datetime): Datetime queued.
            started (datetime): Datetime started.
            ended (datetime): Datetime ended.
            duration (str): Duration.
            exception (str): Exception message if there is a failure.
            trigger (str): Command trigger.
            stateChangeTime (datetime): Datetime the sate was cahnged.
            sendUpdatesToClient (bool): Send Updates To Client.
            updateScheduledTask (bool):Update Scheduled Task.
            id (int): ID of the Command.
    """

    def _load(self, data):
        super()._load(data)
        self.name = self._parse(attrs="name")
        self.commandName = self._parse(attrs="commandName")
        self.message = self._parse(attrs="message")
        self.body = self._parse(attrs="body", value_type="dict")
        self.priority = self._parse(attrs="priority")
        self.status = self._parse(attrs="status")
        self.queued = self._parse(attrs="queued", value_type="date")
        self.started = self._parse(attrs="started", value_type="date")
        self.ended = self._parse(attrs="ended", value_type="date")
        self.duration = self._parse(attrs="duration")
        self.exception = self._parse(attrs="exception")
        self.trigger = self._parse(attrs="trigger")
        self.sendUpdatesToClient = self._parse(attrs="sendUpdatesToClient", value_type="bool")
        self.updateScheduledTask = self._parse(attrs="updateScheduledTask", value_type="bool")
        self.stateChangeTime = self._parse(attrs="stateChangeTime", value_type="date")
        self.lastExecutionTime = self._parse(attrs="lastExecutionTime", value_type="date")
        self.id = self._parse(attrs="id", value_type="int")
        self._finish(self.name)

    def _full_load(self):
        return self._raw.get_command_id(self.id)


class Movie(ReloadObj):
    """ Represents a single Movie.

        Attributes:
            id (int): ID of the Movie.
            title (str): Title of the Movie.
            sortTitle (str): Sort Title of the Movie.
            sizeOnDisk (int): Movie Size On Disk.
            status (str): Status of the Movie.
            overview (str): Overview of the Movie.
            inCinemas (datetime): Date the Movie was in Cinemas.
            physicalRelease (datetime): Date the Movie was Physically Released.
            digitalRelease (datetime): Date the Movie was Digitally Released.
            images (List[:class:`~arrapi.objs.simple.Image`]): List of Images for the Movie.
            website (str): Website of the Movie.
            year (int): Year of the Movie.
            hasFile (bool): If the Movie has a file.
            youTubeTrailerId (str): YouTube Trailer ID for the Movie.
            studio (str): Studio of the Movie.
            path (str): Path of the Movie.
            monitored (bool): If the Movie is monitored.
            minimumAvailability (str): Minimum Availability of the Movie.
            isAvailable (bool): If the Movie is Available.
            folderName (str): Folder Name of the Movie.
            folder (str): Folder name.
            runtime (int): Runtime of the Movie.
            cleanTitle (str): Clean Title of the Movie.
            imdbId (str): IMDb ID of the Movie.
            tmdbId (int): TMDb ID of the Movie.
            titleSlug (str): Title Slug of the Movie.
            certification (str): Certification of the Movie.
            genres (List[str]): List of genres for the Movie.
            tags (List[:class:`~arrapi.objs.reload.Tag`]): List of tags for the Movie.
            added (datetime): Date the Movie was added to Radarr.
            tagsIds (List[int]): List of tag ids for the Movie.
            rating_votes (int): Number of votes for the Movie.
            rating_value (float): Rating of the Movie.
            originalTitle (str): Original Title of the Movie. (Radarr v3 Only)
            digitalRelease (datetime): Date the Movie was Released Digitally. (Radarr v3 Only)
            qualityProfileId (int): Quality Profile ID of the Movie. (Radarr v3 Only)
            qualityProfile (:class:`~arrapi.objs.reload.QualityProfile`): Quality Profile of the Movie. (Radarr v3 Only)
            collection_name (str): Collection Name of the Movie. (Radarr v3 Only)
            collection_tmdbId (int): TMDb Collection ID for the Movie. (Radarr v3 Only)
            downloaded (bool): If the Movie has been Downloaded. (Radarr v2 Only)
            profileId (int): Quality Profile ID of the Movie. (Radarr v2 Only)
            profile (:class:`~arrapi.objs.reload.QualityProfile`): Quality Profile of the Movie. (Radarr v2 Only)
    """

    def __init__(self, radarr, data=None, movie_id=None, tmdb_id=None, imdb_id=None):
        self._loading = True
        self.id = movie_id
        self.tmdbId = tmdb_id
        self.imdbId = imdb_id
        super().__init__(radarr, data)

    def _load(self, data):
        super()._load(data)
        self.title = self._parse(attrs="title")
        self.sortTitle = self._parse(attrs="sortTitle")
        self.sizeOnDisk = self._parse(attrs="sizeOnDisk", value_type="int")
        self.status = self._parse(attrs="status")
        self.overview = self._parse(attrs="overview")
        self.inCinemas = self._parse(attrs="inCinemas", value_type="date")
        self.physicalRelease = self._parse(attrs="physicalRelease", value_type="date")
        self.digitalRelease = self._parse(attrs="digitalRelease", value_type="date")
        self.images = self._parse(attrs="images", value_type="image", is_list=True)
        self.website = self._parse(attrs="website")
        self.year = self._parse(attrs="year", value_type="int")
        self.hasFile = self._parse(attrs="hasFile", value_type="bool")
        self.youTubeTrailerId = self._parse(attrs="youTubeTrailerId")
        self.studio = self._parse(attrs="studio")
        self.path = self._parse(attrs="path")
        self.monitored = self._parse(attrs="monitored", value_type="bool")
        self.minimumAvailability = self._parse(attrs="minimumAvailability")
        self.isAvailable = self._parse(attrs="isAvailable", value_type="bool")
        self.folderName = self._parse(attrs="folderName")
        self.folder = self._parse(attrs="folder")
        self.runtime = self._parse(attrs="runtime", value_type="int")
        self.cleanTitle = self._parse(attrs="cleanTitle")
        self.imdbId = self._parse(attrs="imdbId")
        self.tmdbId = self._parse(attrs="tmdbId", value_type="int", default_is_none=True)
        self.titleSlug = self._parse(attrs="titleSlug")
        self.certification = self._parse(attrs="certification")
        self.genres = self._parse(attrs="genres", is_list=True)
        self.tagsIds = self._parse(attrs="tags", value_type="int", is_list=True)
        self.tags = self._parse(attrs="tags", value_type="intTag", is_list=True)
        self.added = self._parse(attrs="added", value_type="date")
        self.rating_votes = self._parse(attrs=["rating", "votes"], value_type="int")
        self.rating_value = self._parse(attrs=["rating", "value"], value_type="float")
        self.collection = self._parse(attrs="collection", value_type="collection")
        self.id = self._parse(attrs="id", value_type="int", default_is_none=True)
        if self._raw.new_codebase:
            self.originalTitle = self._parse(attrs="originalTitle")
            self.digitalRelease = self._parse(attrs="digitalRelease", value_type="date")
            self.qualityProfileId = self._parse(attrs="qualityProfileId", value_type="int")
            self.qualityProfile = self._parse(attrs="qualityProfileId", value_type="intQualityProfile")
            self.collection_name = self._parse(attrs=["collection", "name"])
            self.collection_tmdbId = self._parse(attrs=["collection", "tmdbId"], value_type="int", default_is_none=True)
        else:
            self.downloaded = self._parse(attrs="downloaded", value_type="bool")
            self.profileId = self._parse(attrs="profileId", value_type="int")
            self.profile = self._parse(attrs="profileId", value_type="intQualityProfile")
        self._finish(self.title)

    def _full_load(self):
        if self.id:
            return self._raw.get_movie_id(self.id)
        elif self.tmdbId or self.imdbId:
            items = self._raw.get_movie_lookup(f"tmdb:{self.tmdbId}" if self.tmdbId else f"imdb:{self.imdbId}")
            if items:
                return items[0]
            else:
                raise NotFound("Item Not Found")
        else:
            raise Invalid("Load Failed: No Load Input")

    def _get_add_data(self, options, path=None):
        if self.id:
            raise Exists(f"{self.title} is already in Radarr")
        self._data.pop("Id", None)
        self._data["monitored"] = options["monitor"]
        if path:
            if not path.startswith(options["root_folder"]):
                raise Invalid(f"Individual Path: {path} must start with the Root Folder: {options['root_folder']} ")
            self._data["path"] = path
        else:
            self._data["rootFolderPath"] = options["root_folder"]
        self._data["rootFolderPath"] = options["root_folder"]
        self._data["qualityProfileId" if self._raw.new_codebase else "profileId"] = options["quality_profile"]
        self._data["minimumAvailability"] = options["minimum_availability"]
        self._data["addOptions"] = {"searchForMovie": options["search"]}
        if "tags" in options:
            self._data["tags"] = options["tags"]
        return self._data

    def add(self,
            root_folder: Union[str, int, "RootFolder"],
            quality_profile: Union[str, int, "QualityProfile"],
            monitor: bool = True,
            search: bool = True,
            minimum_availability: str = "announced",
            tags: Optional[List[Union[str, int, Tag]]] = None
            ) -> None:
        """ Add this Movie to Radarr.

            Parameters:
                root_folder (Union[str, int, RootFolder]): Root Folder for the Movie.
                quality_profile (Union[str, int, QualityProfile]): Quality Profile for the Movie.
                monitor (bool): Monitor the Movie.
                search (bool): Search for the Movie after adding.
                minimum_availability (str): Minimum Availability for the Movie. Valid options are announced, inCinemas, released, or preDB.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added to the Movie.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.Exists`: When the Movie already Exists in Radarr.
        """
        if self._arr.exclusions and self.tmdbId in self._arr.exclusions:
            raise Excluded(f"TMDb ID: {self.tmdbId} is excluded from being added.")
        self._load(self._raw.post_movie(self._get_add_data(self._arr._validate_add_options(
            root_folder,
            quality_profile,
            monitor=monitor,
            search=search,
            minimum_availability=minimum_availability,
            tags=tags
        ))))

    def edit(self,
             path: Optional[str] = None,
             move_files: bool = False,
             quality_profile: Optional[Union[str, int, "QualityProfile"]] = None,
             monitored: Optional[bool] = None,
             minimum_availability: Optional[str] = None,
             tags: Optional[List[Union[str, int, Tag]]] = None,
             apply_tags: str = "add"
             ) -> None:
        """ Edit this Movie in Radarr.

            Parameters:
                path (Optional[str]): Path to change the Movie to.
                move_files (bool): When changing the path do you want to move the files to the new path.
                quality_profile (Optional[Union[str, int, QualityProfile]]): Quality Profile to change the Movie to.
                monitored (Optional[bool]): Monitor the Movie.
                minimum_availability (Optional[str]): Minimum Availability to change the Movie to. Valid options are announced, inCinemas, released, or preDB.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added, replaced, or removed from the Movie.
                apply_tags (str): How you want to edit the Tags. Valid options are add, replace, or remove.

            Raises:
                :class:`ValueError`: When theres no options given.
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.NotFound`: When the Movie hasn't been added to Radarr.
        """
        if not self.id:
            raise NotFound(f"{self.title} not found Radarr, it must be added before editing")
        options = self._arr._validate_edit_options(path=path, move_files=move_files, quality_profile=quality_profile,
                                                   monitored=monitored, minimum_availability=minimum_availability,
                                                   tags=tags, apply_tags=apply_tags)
        valid_move_files = options["path"] if "path" in options else False
        for key, value in options.items():
            if key == "tags":
                tag_type = options["applyTags"]
                if tag_type == "add":
                    self._data[key].extend([t for t in value if t not in self._data["tags"]])
                elif tag_type == "remove":
                    self._data[key] = [t for t in self._data["tags"] if t not in value]
                elif tag_type == "replace":
                    self._data[key] = value
                else:
                    raise Invalid(f"Invalid apply_tags: '{tag_type}' Options: {self._arr.apply_tags_options}")
            elif key != ["applyTags", "moveFiles"]:
                self._data[key] = value
        self._load(self._raw.put_movie_id(self.id, self._data, moveFiles=valid_move_files))

    def delete(self, addImportExclusion: bool = False, deleteFiles: bool = False) -> None:
        """ Delete this Movie from Radarr.

            Parameters:
                addImportExclusion (bool): Add Import Exclusion for this Movie.
                deleteFiles (bool): Delete Files for this Movie.

            Raises:
                :class:`~arrapi.exceptions.NotFound`: When the Movie hasn't been added to Radarr.
        """
        if not self.id:
            raise NotFound(f"{self.title} not found Radarr, it must be added before deleting")
        self._raw.delete_movie_id(self.id, addImportExclusion=addImportExclusion, deleteFiles=deleteFiles)
        self._loading = True
        self.id = None
        self._loading = False


class Series(ReloadObj):
    """ Represents a single Series.

        Attributes:
            id (int): ID of the Series.
            title (str): Title of the Series.
            sortTitle (str): Sort Title of the Series.
            status (str): Status of the Series.
            overview (str): Overview of the Series.
            seasons (List[:class:`~arrapi.objs.simple.Season`]): List of Seasons in the Series.
            nextAiring (datetime): Date the next Episode in the Series Airs.
            previousAiring (datetime): Date the latest Episode in the Series Aired.
            network (str): Network the Series Airs on.
            airTime (str): Time Series Airs.
            images (List[:class:`~arrapi.objs.simple.Image`]): List of Images for the Series.
            year (int): Year of the Series.
            path (str): Path of the Series.
            languageProfileId (int): Language Profile ID of the Series.
            languageProfile (:class:`~arrapi.objs.reload.LanguageProfile`)): Language Profile of the Series.
            seasonFolder (bool): If the Series has Season Folders.
            monitored (bool): If the Series is monitored.
            useSceneNumbering (bool): If the Series uses Scene Numbering.
            folder (str): Folder name.
            runtime (int): Runtime of the Series.
            cleanTitle (str): Clean Title of the Series.
            imdbId (str): IMDb ID of the Series.
            tvdbId (int): TVDb ID of the Series.
            tvRageId (int): TV Rage ID of the Series.
            tvMazeId (int): TV Maze ID of the Series.
            titleSlug (str): Title Slug of the Series.
            firstAired (datetime): Date the Series First Aired.
            seriesType (str): Series Type of the Series.
            certification (str): Certification of the Series.
            genres (List[str]): List of genres for the Series.
            tags (List[:class:`~arrapi.objs.reload.Tag`]): List of tags for the Movie.
            added (datetime): Date the Movie was added to Radarr.
            tagsIds (List[int]): List of tag ids for the Movie.
            rating_votes (int): Number of votes for the Series.
            rating_value (float): Rating of the Series.
            seasonCount (int): Number of Seasons in the Series.
            totalEpisodeCount (int): Total Episodes in the Series.
            episodeCount (int): Episodes in the Series.
            episodeFileCount (int): Episode File Count for the Series.
            sizeOnDisk (int): Size on Disk for the Series.
            ended (bool): If the Series has ended. (Sonarr v3 Only)
            rootFolderPath (str): Root Folder Path for the Series. (Sonarr v3 Only)
            qualityProfileId (int): Quality Profile ID of the Series. (Sonarr v3 Only)
            qualityProfile (:class:`~arrapi.objs.reload.QualityProfile`): Quality Profile of the Series. (Sonarr v3 Only)
            percentOfEpisodes (float): Percent of Episodes obtained. (Sonarr v3 Only)
            profileId (int): Quality Profile ID of the Series. (Sonarr v2 Only)
            profile (:class:`~arrapi.objs.reload.QualityProfile`): Quality Profile of the Series. (Sonarr v2 Only)
    """

    def __init__(self, sonarr, data=None, series_id=None, tvdb_id=None):
        self._loading = True
        self.id = series_id
        self.tvdbId = tvdb_id
        super().__init__(sonarr, data, load=series_id or tvdb_id)

    def _load(self, data):
        super()._load(data)
        self.title = self._parse(attrs="title")
        self.sortTitle = self._parse(attrs="sortTitle")
        self.status = self._parse(attrs="status")
        self.overview = self._parse(attrs="overview")
        self.nextAiring = self._parse(attrs="nextAiring", value_type="date")
        self.previousAiring = self._parse(attrs="previousAiring", value_type="date")
        self.network = self._parse(attrs="network")
        self.airTime = self._parse(attrs="airTime")
        self.images = self._parse(attrs="images", value_type="image", is_list=True)
        self.year = self._parse(attrs="year", value_type="int")
        self.path = self._parse(attrs="path")
        self.languageProfileId = self._parse(attrs="languageProfileId", value_type="int")
        self.languageProfile = self._parse(attrs="languageProfileId", value_type="intLanguageProfile")
        self.seasonFolder = self._parse(attrs="seasonFolder", value_type="bool")
        self.monitored = self._parse(attrs="monitored", value_type="bool")
        self.useSceneNumbering = self._parse(attrs="useSceneNumbering", value_type="bool")
        self.folder = self._parse(attrs="folder")
        self.runtime = self._parse(attrs="runtime", value_type="int")
        self.cleanTitle = self._parse(attrs="cleanTitle")
        self.imdbId = self._parse(attrs="imdbId")
        self.tvdbId = self._parse(attrs="tvdbId", value_type="int")
        self.tvRageId = self._parse(attrs="tvRageId", value_type="int")
        self.tvMazeId = self._parse(attrs="tvMazeId", value_type="int")
        self.titleSlug = self._parse(attrs="titleSlug")
        self.firstAired = self._parse(attrs="firstAired", value_type="date")
        self.seriesType = self._parse(attrs="seriesType")
        self.certification = self._parse(attrs="certification")
        self.genres = self._parse(attrs="genres", is_list=True)
        self.tagsIds = self._parse(attrs="tags", value_type="int", is_list=True)
        self.tags = self._parse(attrs="tags", value_type="intTag", is_list=True)
        self.added = self._parse(attrs="added", value_type="date")
        if "rating" in self._data:
            self.rating_votes = self._parse(attrs=["rating", "votes"], value_type="int")
            self.rating_value = self._parse(attrs=["rating", "value"], value_type="float")
        self.id = self._parse(attrs="id", value_type="int", default_is_none=True)
        self.seasons = self._parse(attrs="seasons", value_type="season", is_list=True)

        if self._raw.new_codebase:
            self.ended = self._parse(attrs="ended", value_type="bool")
            self.rootFolderPath = self._parse(attrs="rootFolderPath")
            self.qualityProfileId = self._parse(attrs="qualityProfileId", value_type="int")
            self.qualityProfile = self._parse(attrs="qualityProfileId", value_type="intQualityProfile")
            self.seasonCount = self._parse(attrs=["statistics", "seasonCount"], value_type="int")
            self.totalEpisodeCount = self._parse(attrs=["statistics", "totalEpisodeCount"], value_type="int")
            self.episodeCount = self._parse(attrs=["statistics", "episodeCount"], value_type="int")
            self.episodeFileCount = self._parse(attrs=["statistics", "episodeFileCount"], value_type="int")
            self.sizeOnDisk = self._parse(attrs=["statistics", "sizeOnDisk"], value_type="int")
            self.percentOfEpisodes = self._parse(attrs=["statistics", "percentOfEpisodes"], value_type="float")
        else:
            self.profileId = self._parse(attrs="profileId", value_type="int")
            self.profile = self._parse(attrs="profileId", value_type="intQualityProfile")
            self.seasonCount = self._parse(attrs="seasonCount", value_type="int")
            self.totalEpisodeCount = self._parse(attrs="totalEpisodeCount", value_type="int")
            self.episodeCount = self._parse(attrs="episodeCount", value_type="int")
            self.episodeFileCount = self._parse(attrs="episodeFileCount", value_type="int")
            self.sizeOnDisk = self._parse(attrs="sizeOnDisk", value_type="int")
        self._finish(self.title)

    def _full_load(self):
        if self.id:
            return self._raw.get_series_id(self.id)
        elif self.tvdbId:
            items = self._raw.get_series_lookup(f"tvdb:{self.tvdbId}")
            if items:
                return items[0]
            else:
                raise NotFound("Item Not Found")
        else:
            raise Invalid("Load Failed: No Load Input")

    def _get_add_data(self, options, path=None):
        if self.id:
            raise Exists(f"{self.title} is already in Sonarr")
        self._data.pop("Id", None)
        self._data["rootFolderPath"] = options["root_folder"]
        if path:
            if not path.startswith(options["root_folder"]):
                raise Invalid(f"Individual Path: {path} must start with the Root Folder: {options['root_folder']} ")
            self._data["path"] = path
        else:
            self._data["rootFolderPath"] = options["root_folder"]
        self._data["monitored"] = options["monitored"]
        self._data["qualityProfileId" if self._raw.new_codebase else "profileId"] = options["quality_profile"]
        if "language_profile" in options:
            self._data["languageProfileId"] = options["language_profile"]
        self._data["seriesType"] = options["series_type"]
        self._data["seasonFolder"] = options["season_folder"]
        self._data["addOptions"] = {
            "searchForMissingEpisodes": options["search"],
            "searchForCutoffUnmetEpisodes": options["unmet_search"],
            "monitor": options["monitor"]
        }
        if "tags" in options:
            self._data["tags"] = options["tags"]
        return self._data

    def add(self,
            root_folder: Union[str, int, "RootFolder"],
            quality_profile: Union[str, int, "QualityProfile"],
            language_profile: Optional[Union[str, int, "LanguageProfile"]] = None,
            monitor: str = "all",
            season_folder: bool = True,
            search: bool = True,
            unmet_search: bool = True,
            series_type: str = "standard",
            tags: Optional[List[Union[str, int, Tag]]] = None) -> None:
        """ Add this Series to Sonarr.

            Parameters:
                root_folder (Union[str, int, RootFolder]): Root Folder for the Series.
                quality_profile (Union[str, int, QualityProfile]): Quality Profile for the Series.
                language_profile (Optional[Union[str, int, LanguageProfile]]): Language Profile for the Series. Required for older versions.
                monitor (bool): How to monitor the Series. Valid options are all, future, missing, existing, pilot, firstSeason, latestSeason, or none.
                season_folder (bool): Use Season Folders for the Series.
                search (bool): Start search for missing episodes of the Series after adding.
                unmet_search (bool): Start search for cutoff unmet episodes of the Series after adding.
                series_type (str): Series Type for the Series. Valid options are standard, daily, or anime.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added to the Series.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.Exists`: When the Series already Exists in Sonarr.
        """
        if self._arr.exclusions and self.tvdbId in self._arr.exclusions:
            raise Excluded(f"TVDb ID: {self.tvdbId} is excluded from being added.")
        self._load(self._raw.post_series(self._get_add_data(self._arr._validate_add_options(
            root_folder,
            quality_profile,
            language_profile=language_profile,
            monitor=monitor,
            season_folder=season_folder,
            search=search,
            unmet_search=unmet_search,
            series_type=series_type,
            tags=tags
        ))))

    def edit(self,
             path: Optional[str] = None,
             move_files: bool = False,
             quality_profile: Optional[Union[str, int, "QualityProfile"]] = None,
             language_profile: Optional[Union[str, int, "LanguageProfile"]] = None,
             monitor: Optional[str] = None,
             monitored: Optional[bool] = None,
             season_folder: Optional[bool] = None,
             series_type: Optional[str] = None,
             tags: Optional[List[Union[str, int, Tag]]] = None,
             apply_tags: str = "add"
             ) -> None:
        """ Edit this Series in Sonarr.

            Parameters:
                path (Optional[str]): Path to change the Series to.
                move_files (bool): When changing the path do you want to move the files to the new path.
                quality_profile (Optional[Union[str, int, QualityProfile]]): Quality Profile to change the Series to.
                language_profile (Optional[Union[str, int, LanguageProfile]]): Language Profile to change the Series to.
                monitor (Optional[str]): How you want the Series monitored. Valid options are all, future, missing, existing, pilot, firstSeason, latestSeason, or none.
                monitored (Optional[bool]): Monitor the Series.
                season_folder (Optional[bool]): Use Season Folders for the Series.
                series_type (Optional[str]): Series Type to change the Series to. Valid options are standard, daily, or anime.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added, replaced, or removed from the Series.
                apply_tags (str): How you want to edit the Tags. Valid options are add, replace, or remove.

            Raises:
                :class:`ValueError`: When theres no options given.
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.NotFound`: When the Series hasn't been added to Sonarr.
        """
        if not self.id:
            raise NotFound(f"{self.title} not found in Sonarr, it must be added before editing")
        options = self._arr._validate_edit_options(path=path, move_files=move_files, quality_profile=quality_profile,
                                                   language_profile=language_profile, monitor=monitor,
                                                   monitored=monitored, season_folder=season_folder,
                                                   series_type=series_type, tags=tags, apply_tags=apply_tags)
        if "monitor" in options:
            self._raw.edit_series_monitoring([self.id], options.pop("monitor"))
        valid_move_files = options["path"] if "path" in options else False
        for key, value in options.items():
            if key == "tags":
                tag_type = options["applyTags"]
                if tag_type == "add":
                    self._data[key].extend([t for t in value if t not in self._data["tags"]])
                elif tag_type == "remove":
                    self._data[key] = [t for t in self._data["tags"] if t not in value]
                elif tag_type == "replace":
                    self._data[key] = value
                else:
                    raise Invalid(f"Invalid apply_tags: '{tag_type}' Options: {self._arr.apply_tags_options}")
            elif key != ["applyTags", "moveFiles"]:
                self._data[key] = value
        self._load(self._raw.put_series_id(self.id, self._data, moveFiles=valid_move_files))

    def delete(self, addImportListExclusion: bool = False, deleteFiles: bool = False) -> None:
        """ Delete this Series from Sonarr.

            Parameters:
                addImportListExclusion (bool): Add Import Exclusion for this Series.
                deleteFiles (bool): Delete Files for this Series.

            Raises:
                :class:`~arrapi.exceptions.NotFound`: When the Series hasn't been added to Sonarr.
        """
        if not self.id:
            raise NotFound(f"{self.title} not found in Sonarr, it must be added before deleting")
        self._raw.delete_series_id(self.id, addImportListExclusion=addImportListExclusion, deleteFiles=deleteFiles)
        self._loading = True
        self.id = None
        self._loading = False



from requests import Session
from typing import Optional
from .base import BaseRawAPI


class SonarrRawAPI(BaseRawAPI):
    def __init__(self, url: str, apikey: str, session: Optional[Session] = None) -> None:
        super().__init__(url, apikey, session=session)

    def get_series(self, tvdb_id=None):
        """ GET /series """
        if tvdb_id is not None:
            return self._get("series", **{"tvdbId": tvdb_id})
        else:
            return self._get("series")

    def get_series_id(self, series_id):
        """ GET /series/{id} """
        return self._get(f"series/{series_id}")

    def post_series(self, json):
        """ POST /series """
        return self._post("series", json=json)

    def post_series_import(self, json):
        """ POST /series/import """
        return self._post("series/import", json=json)

    def put_series(self, json, moveFiles=False):
        """ PUT /series """
        params = {"moveFiles": "true"} if moveFiles else {}
        return self._put("series", json=json, **params)

    def put_series_id(self, series_id, json, moveFiles=False):
        """ PUT /series/{id} """
        params = {"moveFiles": "true"} if moveFiles else {}
        return self._put(f"series/{series_id}", json=json, **params)

    def put_series_editor(self, json):
        """ PUT /series/editor """
        return self._put("series/editor", json=json)

    def delete_series_id(self, series_id, addImportListExclusion=False, deleteFiles=False):
        """ DELETE /series/{id} """
        params = {}
        if addImportListExclusion:
            params["addImportListExclusion"] = "true"
        if deleteFiles:
            params["deleteFiles"] = "true"
        self._delete(f"series/{series_id}", **params)

    def delete_series_editor(self, json):
        """ DELETE /series/editor """
        return self._delete("series/editor", json=json)

    def get_series_lookup(self, term):
        """ GET /series/lookup """
        return self._get("series/lookup", **{"term": term})

    def post_seasonPass(self, json):
        """ POST /seasonPass """
        return self._post("seasonPass", json=json)

    def edit_series_monitoring(self, series_ids, monitor):
        """ Edit multiple Series monitoring """
        if not isinstance(series_ids, list):
            series_ids = [series_ids]
        monitored = monitor != "none"
        json = {
            "monitoringOptions": {"monitor": monitor},
            "series": [{"id": s, "monitored": monitored} for s in series_ids]
        }
        return self.post_seasonPass(json)

    def get_languageProfile(self):
        """ GET /languageProfile """
        return self._get("languageProfile")

    def get_languageProfileId(self, languageProfileId):
        """" GET /languageProfile/languageProfileId """
        return self._get(f"languageProfile/{languageProfileId}")

    def get_importlistexclusion(self):
        """ GET /importlistexclusion """
        return self._get("importlistexclusion")

    def post_importlistexclusion(self, json):
        """ POST /importlistexclusion """
        return self._post("importlistexclusion", json=json)

    def add_importlistexclusion(self, title, tvdb_id):
        return self.post_importlistexclusion({
            "title": title,
            "tvdbId": tvdb_id
        })


from abc import ABC, abstractmethod
from arrapi import Invalid, SystemStatus, QualityProfile, MetadataProfile, RootFolder, Tag, RemotePathMapping
from typing import List

from arrapi.objs.reload import Command


class BaseAPI(ABC):
    @abstractmethod
    def __init__(self, raw):
        self._raw = raw
        self.apply_tags_options = ["add", "remove", "replace"]

    def _validate_options(self, title: str, value: str, options: List[str]):
        """ Validate the value given from the options given.

            Parameters:
                title (str): Name of what is being validated.
                value (str): Value to check options for.
                options (List[str]): List of options to check the value against.

            Returns:
                str: Valid Value

            Raises:
                :class:`~arrapi.exceptions.Invalid`: If the value isn't in the options.
        """
        if value in options:
            return value
        raise Invalid(f"Invalid {title}: '{value}' Options: {options}")

    def _validate_tags(self, tags, create=True):
        """ Checks to see if tags are valid and if create=True will create any tags not found. """
        if not isinstance(tags, list):
            tags = [tags]

        if create is True:
            all_tag_labels = []
            all_tag_ids = []
            for tag in self.all_tags():
                all_tag_labels.append(tag.label)
                all_tag_ids.append(tag.id)
            for tag in tags:
                if not isinstance(tag, (Tag, int)) and str(tag).lower() not in all_tag_labels:
                    self._raw.post_tag(str(tag).lower())

        all_tag_labels = {}
        all_tag_ids = []
        valid_tag_ids = []
        for tag in self.all_tags():
            all_tag_labels[tag.label] = tag.id
            all_tag_ids.append(tag.id)
        for tag in tags:
            if isinstance(tag, Tag) and tag.id in all_tag_ids:
                valid_tag_ids.append(tag.id)
            elif isinstance(tag, int) and tag in all_tag_ids:
                valid_tag_ids.append(tag)
            elif str(tag).lower() in all_tag_labels:
                valid_tag_ids.append(all_tag_labels[str(tag).lower()])

        return valid_tag_ids

    def _validate_apply_tags(self, apply_tags):
        """ Validate Apply Tags options. """
        return self._validate_options("Apply Tags", apply_tags, self.apply_tags_options)

    def get_tag(self, tag_id: int, detail: bool = False) -> Tag:
        """ Get a :class:`~arrapi.objs.reload.Tag` by its ID.

            Parameters:
                tag_id (int): ID of Tag to get.
                detail (bool): Get Tag with details.

            Returns:
                :class:`~arrapi.objs.reload.Tag`: Tag of the ID given.

            Raises:
                :class:`~arrapi.exceptions.NotFound`: When there's no tag with that ID.
        """
        return Tag(self, self._raw.get_tag_id(tag_id, detail=detail))

    def all_tags(self, detail: bool = False) -> List[Tag]:
        """ Gets every :class:`~arrapi.objs.reload.Tag`.

            Parameters:
                detail (bool): Get Tags with details.

            Returns:
                List[:class:`~arrapi.objs.reload.Tag`]: List of all Tags.
        """
        return [Tag(self, data) for data in self._raw.get_tag(detail=detail)]

    def create_tag(self, label: str) -> Tag:
        """ Create a new :class:`~arrapi.objs.reload.Tag`.

            Parameters:
                label (str): Label of new Tag.

            Returns:
                :class:`~arrapi.objs.reload.Tag`: Tag just created.
        """
        return Tag(self, self._raw.post_tag(label))

    def edit_tag(self, tag_id: int, label: str) -> Tag:
        """ Edit a :class:`~arrapi.objs.reload.Tag` by its ID.

            Parameters:
                tag_id (int): ID of tag to edit.
                label (str): Label to change tag to.

            Returns:
                :class:`~arrapi.objs.reload.Tag`: Tag just edited.

            Raises:
                :class:`~arrapi.exceptions.NotFound`: When there's no tag with that ID.
        """
        return Tag(self, self._raw.put_tag_id(tag_id, label))

    def delete_tag(self, tag_id: int) -> None:
        """ Delete a :class:`~arrapi.objs.reload.Tag` by its ID.

            Parameters:
                tag_id (int): ID of tag to delete.

            Raises:
                :class:`~arrapi.exceptions.NotFound`: When there's no tag with that ID.
        """
        self._raw.delete_tag_id(tag_id)

    def all_commands(self) -> List[Command]:
        """ Gets a list of :class:`~arrapi.objs.reload.Command`.

            Returns:
                List[:class:`~arrapi.objs.reload.Command`]: List of Commands
        """
        return [Command(self, data) for data in self._raw.get_command()]

    def get_command(self, command_id: int) -> Command:
        """  Get a :class:`~arrapi.objs.reload.Command` by its ID.

            Parameters:
                command_id (int): ID of command.

            Returns:
                :class:`~arrapi.objs.reload.Command`: Command of the ID given.

            Raises:
                :class:`~arrapi.exceptions.NotFound`: When there's no command with that ID.
        """
        return Command(self, {"id": command_id}, load=True)

    def send_command(self, command: str, **kwargs) -> Command:
        """  Sends a command.

            Parameters:
                command (str): name of command to send.
                kwargs (dict): additional json parameters.

            Returns:
                :class:`~arrapi.objs.reload.Command`: Command created.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When the command given is invalid.
        """
        return Command(self, self._raw.post_command(command, **kwargs))

    def _validate_quality_profile(self, quality_profile):
        """ Validate Quality Profile options. """
        options = []
        for profile in self.quality_profile():
            options.append(profile)
            if (isinstance(quality_profile, QualityProfile) and profile.id == quality_profile.id) \
                    or (isinstance(quality_profile, int) and profile.id == quality_profile) \
                    or profile.name == quality_profile:
                return profile.id
        raise Invalid(f"Invalid Quality Profile: '{quality_profile}' Options: {options}")

    def quality_profile(self) -> List[QualityProfile]:
        """ Gets every :class:`~arrapi.objs.simple.QualityProfile`.

            Returns:
                List[:class:`~arrapi.objs.simple.QualityProfile`]: List of all Quality Profiles
        """
        return [QualityProfile(self, data) for data in self._raw.get_qualityProfile()]

    def _validate_root_folder(self, root_folder):
        """ Validate Root Folder options. """
        options = []
        for folder in self.root_folder():
            options.append(folder)
            if (isinstance(root_folder, RootFolder) and folder.id == root_folder.id) \
                    or (isinstance(root_folder, int) and folder.id == root_folder) \
                    or folder.path == root_folder:
                return folder.path
        raise Invalid(f"Invalid Root Folder: '{root_folder}' Options: {options}")

    def root_folder(self) -> List[RootFolder]:
        """ Gets every :class:`~arrapi.objs.simple.RootFolder`.

            Returns:
                List[:class:`~arrapi.objs.simple.RootFolder`]: List of all Root Folders.
        """
        return [RootFolder(self, data) for data in self._raw.get_rootFolder()]

    def add_root_folder(self, path):
        """ Adds the path given as a root folder

            Parameters:
                path (str): path of new root folder

            Raises:
                :class:`~arrapi.exceptions.ArrException`: When the path does not exist or is already a root folder
        """
        self._raw.add_rootFolder(path)

    def remote_path_mapping(self) -> List[RemotePathMapping]:
        """ Gets every :class:`~arrapi.objs.simple.RemotePathMapping`.

            Returns:
                List[:class:`~arrapi.objs.simple.RemotePathMapping`]: List of all Remote Path Mappings.
        """
        return [RemotePathMapping(self, data) for data in self._raw.get_remotePathMapping()]

    def system_status(self) -> SystemStatus:
        """ Gets the :class:`~arrapi.objs.relaod.SystemStatus`.

            Returns:
                :class:`~arrapi.objs.reload.SystemStatus`: System Status Information.
        """
        return SystemStatus(self)


class BaseV1API(BaseAPI):
    """ Base class for :class:`~arrapi.apis.lidarr.LidarrAPI` and :class:`~arrapi.apis.readarr.ReadarrAPI`
    containing API calls that are identical between Sonarr and Radarr. """

    @abstractmethod
    def __init__(self, raw):
        super().__init__(raw)

    def _validate_metadata_profile(self, metadata_profile):
        """ Validate Metadata Profile options. """
        options = []
        for profile in self.metadata_profile():
            options.append(profile)
            if (isinstance(metadata_profile, MetadataProfile) and profile.id == metadata_profile.id) \
                    or (isinstance(metadata_profile, int) and profile.id == metadata_profile) \
                    or profile.name == metadata_profile:
                return profile.id
        raise Invalid(f"Invalid Metadata Profile: '{metadata_profile}' Options: {options}")

    def metadata_profile(self) -> List[MetadataProfile]:
        """ Gets every :class:`~arrapi.objs.MetadataProfile`.

            Returns:
                List[:class:`~arrapi.objs.MetadataProfile`]: List of all Metadata Profiles
        """
        return [MetadataProfile(self, data) for data in self._raw.get_metadataProfile()]



from requests import Session
from typing import Optional, Union, List, Tuple
from arrapi import RootFolder, QualityProfile, Movie, Tag, NotFound, Invalid, Exists
from .base import BaseAPI
from ..exceptions import Excluded
from ..objs.simple import RadarrExclusion
from ..raws.radarr import RadarrRawAPI


class RadarrAPI(BaseAPI):
    """ Primary Class to use when connecting with the Radarr API

        Parameters:
            url (str): URL of Radarr application.
            apikey (str): apikey for the Radarr application.
            session (Optional[Session]): Session object to use.
     """

    def __init__(self, url: str, apikey: str, session: Optional[Session] = None) -> None:
        super().__init__(RadarrRawAPI(url, apikey, session=session))
        self.exclusions = []
        self.minimum_availability_options = ["announced", "inCinemas", "released", "preDB"]
        self.monitor_options = ["movieOnly", "movieAndCollection", "none"]

    def _validate_add_options(self, root_folder, quality_profile, monitor=True, search=True,
                              minimum_availability="announced", tags=None):
        """ Validate Add Movie options. """
        options = {
            "root_folder": self._validate_root_folder(root_folder),
            "quality_profile": self._validate_quality_profile(quality_profile),
            "monitor": self._validate_monitor(monitor),
            "search": True if search is True else False,
            "minimum_availability": self._validate_minimum_availability(minimum_availability)
        }
        if tags:
            options["tags"] = self._validate_tags(tags)
        return options

    def _validate_edit_options(self, root_folder=None, path=None, move_files=False, quality_profile=None,
                               monitored=None, minimum_availability=None, tags=None, apply_tags="add"):
        """ Validate Edit Movie options. """
        if all(v is None for v in [root_folder, path, quality_profile, monitored, minimum_availability, tags]):
            raise ValueError("Expected either root_folder, path, quality_profile, "
                             "monitored, minimum_availability, or tags args")
        options = {"moveFiles": True if move_files is True else False}
        if root_folder is not None:
            options["rootFolderPath"] = self._validate_root_folder(root_folder)
        if path is not None:
            options["path"] = path
        if quality_profile is not None:
            options["qualityProfileId" if self._raw.new_codebase else "profileId"] = self._validate_quality_profile(quality_profile)
        if monitored is not None:
            options["monitored"] = self._validate_monitor(monitored, name="Monitored")
        if minimum_availability is not None:
            options["minimumAvailability"] = self._validate_minimum_availability(minimum_availability)
        if tags is not None:
            options["tags"] = self._validate_tags(tags, create=apply_tags != "remove")
            if apply_tags in self.apply_tags_options:
                options["applyTags"] = apply_tags
            else:
                raise Invalid(f"Invalid apply_tags: '{apply_tags}' Options: {self.apply_tags_options}")
        return options

    def _validate_minimum_availability(self, minimum_availability):
        """ Validate Minimum Availability options. """
        return self._validate_options("Minimum Availability", minimum_availability, self.minimum_availability_options)

    def _validate_monitor(self, monitor, name="Monitor"):
        """ Validate Monitor options. """
        if self._raw.v4:
            if isinstance(monitor, bool):
                return monitor
            return self._validate_options(name, monitor, self.monitor_options)
        else:
            return True if monitor is True else False

    def _validate_ids(self, ids):
        """ Validate IDs. """
        valid_ids = []
        invalid_ids = []
        used_ids = []
        radarr_ids = {}
        for m in self.all_movies():
            radarr_ids[m.tmdbId] = m
            radarr_ids[str(m.tmdbId)] = m
            radarr_ids[m.imdbId] = m
        for _id in ids:
            if isinstance(_id, Movie) and str(_id.tmdbId) not in used_ids and str(_id.imdbId) not in used_ids:
                valid_ids.append(_id.id)
                used_ids.append(str(_id.tmdbId))
                used_ids.append(str(_id.imdbId))
            elif _id in radarr_ids and str(_id) not in used_ids:
                valid_ids.append(radarr_ids[_id].id)
                used_ids.append(str(_id))
            else:
                invalid_ids.append(_id)
        return valid_ids, invalid_ids

    def respect_list_exclusions_when_adding(self):
        """ Stores all List Exclusions so whenever :func:`~arrapi.objs.reload.Movie.add` or :func:`~arrapi.apis.sonarr.RadarrAPI.add_multiple_movies` is called the additions will be checked against the Exclusion List  """
        self.exclusions = [RadarrExclusion(self, ex).tmdbId for ex in self._raw.get_exclusions()]

    def get_movie(self, movie_id: Optional[int] = None, tmdb_id: Optional[int] = None, imdb_id: Optional[str] = None) -> Movie:
        """ Gets a :class:`~arrapi.objs.reload.Movie` by one of the IDs.

            Parameters:
                movie_id (Optional[int]): Search by Radarr Movie ID.
                tmdb_id (Optional[int]): Search by TMDb ID.
                imdb_id (Optional[int]): Search by IMDb ID.

            Returns:
                :class:`~arrapi.objs.reload.Movie`: Movie for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given.
                :class:`~arrapi.exceptions.NotFound`: When there's no movie with that ID.
        """
        if all(v is None for v in [movie_id, tmdb_id, imdb_id]):
            raise ValueError("Expected either movie_id, tmdb_id or imdb_id args")
        return Movie(self, movie_id=movie_id, tmdb_id=tmdb_id, imdb_id=imdb_id)

    def all_movies(self) -> List[Movie]:
        """ Gets all :class:`~arrapi.objs.reload.Movie` in Radarr.

            Returns:
                List[:class:`~arrapi.objs.reload.Movie`]: List of Movies in Radarr.
        """
        return [Movie(self, data=d) for d in self._raw.get_movie()]

    def search_movies(self, term: str) -> List[Movie]:
        """ Gets a list of :class:`~arrapi.objs.reload.Movie` by a search term.

            Parameters:
                term (str): Term to Search for.

            Returns:
                List[:class:`~arrapi.objs.reload.Movie`]: List of Movie's found.
        """
        return [Movie(self, data=d) for d in self._raw.get_movie_lookup(term)]

    def add_movie(
            self,
            root_folder: Union[str, int, "RootFolder"],
            quality_profile: Union[str, int, "QualityProfile"],
            movie_id: Optional[int] = None,
            tmdb_id: Optional[int] = None,
            imdb_id: Optional[str] = None,
            monitor: Union[bool, str] = True,
            search: bool = True,
            minimum_availability: str = "announced",
            tags: Optional[List[Union[str, int, Tag]]] = None
            ) -> Movie:
        """ Gets a :class:`~arrapi.objs.reload.Movie` by one of the IDs and adds it to Radarr.

            Parameters:
                root_folder (Union[str, int, RootFolder]): Root Folder for the Movie.
                quality_profile (Union[str, int, QualityProfile]): Quality Profile for the Movie.
                movie_id (Optional[int]): Search by Radarr Movie ID.
                tmdb_id (Optional[int]): Search by TMDb ID.
                imdb_id (Optional[int]): Search by IMDb ID.
                monitor (Union[bool, str]): Monitor the Movie. Valid V4 Options are movieOnly, movieAndCollection, none.
                search (bool): Search for the Movie after adding.
                minimum_availability (str): Minimum Availability for the Movie. Valid options are announced, inCinemas, released, or preDB.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added to the Movie.

            Returns:
                :class:`~arrapi.objs.reload.Movie`: Movie for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given.
                :class:`~arrapi.exceptions.NotFound`: When there's no movie with that ID.
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.Exists`: When the Movie already Exists in Radarr.
                :class:`~arrapi.exceptions.Excluded`: When the Movie Exists in the Exclusion List.
        """
        movie = self.get_movie(movie_id=movie_id, tmdb_id=tmdb_id, imdb_id=imdb_id)
        movie.add(root_folder, quality_profile, monitor=monitor, search=search,
                  minimum_availability=minimum_availability, tags=tags)
        return movie

    def edit_movie(
            self,
            movie_id: Optional[int] = None,
            tmdb_id: Optional[int] = None,
            imdb_id: Optional[str] = None,
            path: Optional[str] = None,
            move_files: bool = False,
            quality_profile: Optional[Union[str, int, "QualityProfile"]] = None,
            monitored: Optional[Union[bool, str]] = None,
            minimum_availability: Optional[str] = None,
            tags: Optional[List[Union[str, int, Tag]]] = None,
            apply_tags: str = "add"
            ) -> Movie:
        """ Gets a :class:`~arrapi.objs.reload.Movie` by one of the IDs and edits it in Radarr.

            Parameters:
                movie_id (Optional[int]): Search by Radarr Movie ID.
                tmdb_id (Optional[int]): Search by TMDb ID.
                imdb_id (Optional[int]): Search by IMDb ID.
                path (Optional[str]): Path to change the Movie to.
                move_files (bool): When changing the path do you want to move the files to the new path.
                quality_profile (Optional[Union[str, int, QualityProfile]]): Quality Profile to change the Movie to.
                monitored (Optional[Union[bool, str]]): Monitor the Movie. Valid V4 Options are movieOnly, movieAndCollection, none.
                minimum_availability (Optional[str]): Minimum Availability to change the Movie to. Valid options are announced, inCinemas, released, or preDB.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added, replaced, or removed from the Movie.
                apply_tags (str): How you want to edit the Tags. Valid options are add, replace, or remove.

            Raises:
                :class:`ValueError`: When no ID is given or when there's no options given.
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.NotFound`: When there's no movie with that ID or when the Movie hasn't been added to Radarr.
        """
        movie = self.get_movie(movie_id=movie_id, tmdb_id=tmdb_id, imdb_id=imdb_id)
        movie.edit(path=path, move_files=move_files, quality_profile=quality_profile, monitored=monitored,
                   minimum_availability=minimum_availability, tags=tags, apply_tags=apply_tags)
        return movie

    def delete_movie(
            self,
            movie_id: Optional[int] = None,
            tmdb_id: Optional[int] = None,
            imdb_id: Optional[str] = None,
            addImportExclusion: bool = False,
            deleteFiles: bool = False
    ) -> Movie:
        """ Gets a :class:`~arrapi.objs.reload.Movie` by one of the IDs and deletes it from Radarr.

            Parameters:
                movie_id (Optional[int]): Search by Radarr Movie ID.
                tmdb_id (Optional[int]): Search by TMDb ID.
                imdb_id (Optional[int]): Search by IMDb ID.
                addImportExclusion (bool): Add Import Exclusion for this Movie.
                deleteFiles (bool): Delete Files for this Movie.

            Returns:
                :class:`~arrapi.objs.reload.Movie`: Movie for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given.
                :class:`~arrapi.exceptions.NotFound`: When there's no movie with that ID or when the Movie hasn't been added to Radarr.
        """
        movie = self.get_movie(movie_id=movie_id, tmdb_id=tmdb_id, imdb_id=imdb_id)
        movie.delete(addImportExclusion=addImportExclusion, deleteFiles=deleteFiles)
        return movie

    def add_multiple_movies(self, ids: List[Union[int, str, Movie, Tuple[Union[int, str, Movie], str]]],
                            root_folder: Union[str, int, RootFolder],
                            quality_profile: Union[str, int, QualityProfile],
                            monitor: bool = True,
                            search: bool = True,
                            minimum_availability: str = "announced",
                            tags: Optional[List[Union[str, int, Tag]]] = None,
                            per_request: int = None
                            ) -> Tuple[List[Movie], List[Movie], List[Union[int, str, Movie]], List[int]]:
        """ Adds multiple Movies to Radarr in a single call by their TMDb IDs.

            You can specify the path for each TMDb ID using a tuple in the list instead of just the ID ex. ``(11, "/media/Star Wars (1977)/")``

            The path provided must begin with the root_folder specified.

            Parameters:
                ids (List[Union[int, str, Movie, Tuple[Union[int, str, Movie], str]]]): List of TMDB IDs, IMDb IDs, or Movie lookups to add.
                root_folder (Union[str, int, RootFolder]): Root Folder for the Movies.
                quality_profile (Union[str, int, QualityProfile]): Quality Profile for the Movies.
                monitor (bool): Monitor the Movies.
                search (bool): Search for the Movies after adding.
                minimum_availability (str): Minimum Availability for the Movies. Valid options are announced, inCinemas, released, or preDB.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added to the Movies.
                per_request (int): Number of Movies to add per request.

            Returns:
                Tuple[List[:class:`~arrapi.objs.reload.Movie`], List[:class:`~arrapi.objs.reload.Movie`], List[Union[int, str, Movie]], List[int]]: List of Movies that were able to be added, List of Movies already in Radarr, List of Movies that could not be found, List of Movies that were excluded.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
        """
        options = self._validate_add_options(root_folder, quality_profile, monitor=monitor, search=search,
                                             minimum_availability=minimum_availability, tags=tags)
        json = []
        movies = []
        existing_movies = []
        invalid_ids = []
        excluded_ids = []
        used_ids = []
        for input_item in ids:
            path = input_item[1] if isinstance(input_item, tuple) else None
            item = input_item[0] if isinstance(input_item, tuple) else input_item
            try:
                if isinstance(item, Movie):
                    movie = item
                elif str(item).startswith("tt"):
                    movie = self.get_movie(imdb_id=item)
                else:
                    if int(item) in used_ids or (self.exclusions and int(item) in self.exclusions):
                        raise Excluded(int(item))
                    movie = self.get_movie(tmdb_id=item)
                if movie.tmdbId in used_ids or (self.exclusions and movie.tmdbId in self.exclusions):
                    raise Excluded(movie.tmdbId)
                used_ids.append(movie.tmdbId)
                try:
                    json.append(movie._get_add_data(options, path=path))
                except Exists:
                    existing_movies.append(movie)
            except NotFound:
                invalid_ids.append(input_item)
            except Excluded as e:
                excluded_ids.append(int(str(e)))
        if len(json) > 0:
            if per_request is None:
                per_request = len(json)
            for i in range(0, len(json), per_request):
                movies.extend([Movie(self, data=m) for m in self._raw.post_movie_import(json[i:i+per_request])])
        return movies, existing_movies, invalid_ids, excluded_ids

    def edit_multiple_movies(self, ids: List[Union[int, str, Movie]],
                             root_folder: Optional[Union[str, int, RootFolder]] = None,
                             move_files: bool = False,
                             quality_profile: Optional[Union[str, int, QualityProfile]] = None,
                             monitored: Optional[bool] = None,
                             minimum_availability: Optional[str] = None,
                             tags: Optional[List[Union[str, int, Tag]]] = None,
                             apply_tags: str = "add",
                             per_request: int = None
                             ) -> Tuple[List[Movie], List[Union[int, str, Movie]]]:
        """ Edit multiple Movies in Radarr by their TMDb IDs.

            Parameters:
                ids (List[Union[int, str, Movie]]): List of TMDb IDs, IMDb IDs, or Movie objects you want to edit.
                root_folder (Union[str, int, RootFolder]): Root Folder to change the Movie to.
                move_files (bool): When changing the root folder do you want to move the files to the new path.
                quality_profile (Optional[Union[str, int, QualityProfile]]): Quality Profile to change the Movie to.
                monitored (Optional[bool]): Monitor the Movie.
                minimum_availability (Optional[str]): Minimum Availability to change the Movie to. Valid options are announced, inCinemas, released, or preDB.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added, replaced, or removed from the Movie.
                apply_tags (str): How you want to edit the Tags. Valid options are add, replace, or remove.
                per_request (int): Number of Movies to edit per request.

            Returns:
                Tuple[List[:class:`~arrapi.objs.reload.Movie`], List[Union[int, str, Movie]]]: List of Movies that were able to be edited, List of Movies that could not be found in Radarr.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
        """
        json = self._validate_edit_options(root_folder=root_folder, move_files=move_files,
                                           quality_profile=quality_profile, monitored=monitored,
                                           minimum_availability=minimum_availability, tags=tags, apply_tags=apply_tags)
        movie_list = []
        valid_ids, invalid_ids = self._validate_ids(ids)
        if len(valid_ids) > 0:
            if per_request is None:
                per_request = len(valid_ids)
            for i in range(0, len(valid_ids), per_request):
                json["movieIds"] = valid_ids[i:i+per_request]
                movie_list = [Movie(self, data=m) for m in self._raw.put_movie_editor(json)]
        return movie_list, invalid_ids

    def delete_multiple_movies(self, ids: List[Union[int, str, Movie]],
                               addImportExclusion: bool = False,
                               deleteFiles: bool = False,
                               per_request: int = None
                               ) -> List[Union[int, str, Movie]]:
        """ Deletes multiple Movies in Radarr by their TMDb IDs.

            Parameters:
                ids (List[Union[int, str, Movie]]): List of TMDb IDs, IMDb IDs, or Movie objects you want to delete.
                addImportExclusion (bool): Add Import Exclusion for these TMDb IDs.
                deleteFiles (bool): Delete Files for these TMDb IDs.
                per_request (int): Number of Movies to delete per request.

            Returns:
                List[Union[int, str, Movie]]: List of Movies that could not be found in Radarr.
        """
        valid_ids, invalid_ids = self._validate_ids(ids)
        if len(valid_ids) > 0:
            json = {
                "deleteFiles": deleteFiles,
                "addImportExclusion": addImportExclusion
            }
            if per_request is None:
                per_request = len(valid_ids)
            for i in range(0, len(valid_ids), per_request):
                json["movieIds"] = valid_ids[i:i+per_request]
                self._raw.delete_movie_editor(json)
        return invalid_ids



import logging

from abc import ABC, abstractmethod
from json.decoder import JSONDecodeError
from requests import Session
from requests.exceptions import RequestException
from arrapi import ArrException, ConnectionFailure, NotFound, Unauthorized, Invalid

logger = logging.getLogger(__name__)

class BaseRawAPI(ABC):

    @abstractmethod
    def __init__(self, url, apikey, v1=False, session=None):
        self.url = url.rstrip("/")
        self.apikey = apikey
        self.session = Session() if session is None else session
        self.v1 = v1
        self.v3 = True
        self.v4 = False
        try:
            status = self.get_system_status()
        except NotFound:
            self.v3 = False
            status = self.get_system_status()
        if "version" not in status or status["version"] is None:
            raise ConnectionFailure(f"Failed to Connect to {self.url}")
        if v1 is False:
            major = int(status["version"].split(".")[0])
            self.v3 = major == 3
            self.v4 = major > 3
        self.new_codebase = self.v1 or self.v3 or self.v4

    def _get(self, path, **kwargs):
        """ process get request. """
        return self._request("get", path, **kwargs)

    def _delete(self, path, json=None, **kwargs):
        """ process delete request. """
        return self._request("delete", path, json=json, **kwargs)

    def _post(self, path, json=None, **kwargs):
        """ process post request. """
        return self._request("post", path, json=json, **kwargs)

    def _put(self, path, json=None, **kwargs):
        """ process put request. """
        return self._request("put", path, json=json, **kwargs)

    def _request(self, request_type, path, json=None, **kwargs):
        """ process request. """
        url_params = {"apikey": f"{self.apikey}"}
        for kwarg in kwargs:
            url_params[kwarg] = kwargs[kwarg]
        request_url = f"{self.url}/api{'/v1' if self.v1 else '/v3' if self.v3 or self.v4 else ''}/{path}"
        if json is not None:
            logger.debug(f"Request JSON {json}")
        try:
            if request_type == "delete":
                response = self.session.delete(request_url, json=json, params=url_params)
            elif request_type == "post":
                response = self.session.post(request_url, json=json, params=url_params)
            elif request_type == "put":
                response = self.session.put(request_url, json=json, params=url_params)
            else:
                response = self.session.get(request_url, params=url_params)
        except RequestException:
            raise ConnectionFailure(f"Failed to Connect to {self.url}")
        try:
            response_json = response.json()
        except JSONDecodeError:
            logger.debug(f"Response ({response.status_code} [{response.reason}]) {response.content}")
            if response.status_code >= 400:
                raise ArrException(f"({response.status_code} [{response.reason}]) {response.content}")
            else:
                return None
        else:
            logger.debug(f"Response ({response.status_code} [{response.reason}]) {response_json}")
            if response.status_code == 401:
                raise Unauthorized(f"({response.status_code} [{response.reason}]) Invalid API Key {response_json}")
            elif response.status_code == 404:
                raise NotFound(f"({response.status_code} [{response.reason}]) Item Not Found {response_json}")
            elif response.status_code == 500 and "message" in response_json and response_json["message"] == "Sequence contains no matching element":
                raise Invalid(f"({response.status_code} [{response.reason}]) Invalid option provided")
            elif response.status_code >= 400:
                if isinstance(response_json, list) and "errorMessage" in response_json[0]:
                    raise ArrException(f"({response.status_code} [{response.reason}]) {response_json[0]['errorMessage']}")
                else:
                    raise ArrException(f"({response.status_code} [{response.reason}]) {response_json}")
            return response_json

    def get_tag(self, detail=False):
        """ GET /tag and GET /tag/detail """
        return self._get("tag/detail" if detail and self.new_codebase else "tag")

    def post_tag(self, label):
        """ POST /tag """
        return self._post("tag", json={"label": str(label).lower()})

    def get_tag_id(self, tag_id, detail=False):
        """ GET /tag/{id} and GET /tag/detail/{id} """
        return self._get(f"tag/detail/{tag_id}" if detail and self.new_codebase else f"tag/{tag_id}")

    def put_tag_id(self, tag_id, label):
        """ PUT /tag/{id} """
        return self._put(f"tag/{tag_id}", json={"id": tag_id, "label": str(label).lower()})

    def delete_tag_id(self, tag_id):
        """ DELETE /tag/{id} """
        return self._delete(f"tag/{tag_id}")

    def get_command(self):
        """ GET /command """
        return self._get("command")

    def get_command_id(self, command_id):
        """ GET /command/{id} """
        return self._get(f"command/{command_id}")

    def post_command(self, command, **kwargs):
        """ POST /command """
        json = {k: v for k, v in kwargs.items()}
        json["name"] = command
        return self._post("command", json=json)

    def get_qualityProfile(self):
        """" GET /qualityProfile for v3 and GET /profile for v2 """
        return self._get("qualityProfile" if self.new_codebase else "profile")

    def get_qualityProfileId(self, qualityProfileId):
        """" GET /qualityProfile/qualityProfileId for v3 and GET /profile/qualityProfileId for v2 """
        return self._get(f"qualityProfile/{qualityProfileId}" if self.new_codebase else f"profile/{qualityProfileId}")

    def get_rootFolder(self):
        """ GET /rootFolder """
        return self._get("rootFolder")

    def post_rootFolder(self, json):
        """ POST /rootFolder """
        return self._post("rootFolder", json=json)

    def add_rootFolder(self, path):
        return self.post_rootFolder({"path": path})

    def delete_rootFolder(self, rootFolderID):
        self._delete(f"rootFolder/{rootFolderID}")

    def get_remotePathMapping(self):
        """ GET /remotePathMapping """
        return self._get("remotePathMapping")

    def get_system_status(self):
        """ GET /system/status """
        return self._get("system/status")


class BaseRawV1API(BaseRawAPI):

    @abstractmethod
    def __init__(self, url, apikey, session=None):
        super().__init__(url, apikey, v1=True, session=session)

    def get_metadataProfile(self):
        """ GET /metadataProfile """
        return self._get("metadataProfile")


from abc import abstractmethod

from arrapi.objs.base import BaseObj

class SimpleObj(BaseObj):
    @abstractmethod
    def _load(self, data):
        super()._load(data)


class Collection(SimpleObj):
    """ Represents a single Collection.

        Attributes:
            name (str): Name of the Collection.
            tmdbId (int): TMDb Collection ID of the Collection.
    """

    def _load(self, data):
        super()._load(data)
        self.name = self._parse(attrs="name")
        self.tmdbId = self._parse(attrs="tmdbId", value_type="int")
        self._finish(self.name)

    def get_url(self):
        """ URL of the Collection on TMDb"""
        return f"https://www.themoviedb.org/collection/{self.tmdbId}"


class Image(SimpleObj):
    """ Represents a single Image.

        Attributes:
            coverType (str): Cover Type of the Image.
            url (str): URL of the Image.
            remoteUrl (str): Remote URL of the Image.
    """

    def _load(self, data):
        super()._load(data)
        self.coverType = self._parse(attrs="coverType")
        self.url = self._parse(attrs="url")
        self.remoteUrl = self._parse(attrs="remoteUrl")
        self._finish(self.remoteUrl)


class MetadataProfile(SimpleObj):
    """ Represents a single Metadata Profile.

        Attributes:
            id (int): ID of the Metadata Profile.
            name (str): Name of the Metadata Profile.
    """

    def _load(self, data):
        super()._load(data)
        self.id = self._parse(attrs="id", value_type="int")
        self.name = self._parse(attrs="name")
        self._finish(self.name)

class RemotePathMapping(SimpleObj):
    """ Represents a single Remote Path Mapping.

        Attributes:
            id (int): ID of the Remote Path Mapping.
            host (str): Host of the Remote Path Mapping.
            localPath (str): Local Path of the Remote Path Mapping.
            remotePath (str): Remote Path of the Remote Path Mapping.
    """

    def _load(self, data):
        super()._load(data)
        self.id = self._parse(attrs="id", value_type="int")
        self.host = self._parse(attrs="host")
        self.remotePath = self._parse(attrs="remotePath")
        self.localPath = self._parse(attrs="localPath")
        self._finish(self.host)

class RootFolder(SimpleObj):
    """ Represents a single Root Folder.

        Attributes:
            id (int): ID of the Root Folder.
            path (str): Path of the Root Folder.
            freeSpace (int): Free Space in the Root Folder.
            name (str): Name of the Root Folder. (Only when loaded using :class:`~arrapi.apis.lidarr.LidarrAPI` or :class:`~arrapi.apis.readarr.ReadarrAPI`)
            defaultMetadataProfileId (int): Default Metadata Profile ID of the Root Folder. (Only when loaded using :class:`~arrapi.apis.lidarr.LidarrAPI` or :class:`~arrapi.apis.readarr.ReadarrAPI`)
            defaultQualityProfileId (int): Default Quality Profile ID of the Root Folder. (Only when loaded using :class:`~arrapi.apis.lidarr.LidarrAPI` or :class:`~arrapi.apis.readarr.ReadarrAPI`)
            defaultMonitorOption (int): Default Monitor Option of the Root Folder. (Only when loaded using :class:`~arrapi.apis.lidarr.LidarrAPI` or :class:`~arrapi.apis.readarr.ReadarrAPI`)
            defaultTags (int): Default Tags of the Root Folder. (Only when loaded using :class:`~arrapi.apis.lidarr.LidarrAPI` or :class:`~arrapi.apis.readarr.ReadarrAPI`)
            isCalibreLibrary (bool): If the Root Folder is a Calibre Library. (Only when loaded using :class:`~arrapi.apis.readarr.ReadarrAPI`)
            unmappedFolders (List[UnmappedFolder]): Unmapped Folders in the Root Folder. (Only when loaded using :class:`~arrapi.apis.radarr.SonarrAPI` V3 or :class:`~arrapi.apis.radarr.RadarrAPI` V3)
    """

    def _load(self, data):
        super()._load(data)
        self.path = self._parse(attrs="path")
        self.id = self._parse(attrs="id", value_type="int")
        if "name" in self._data:
            self.name = self._parse(attrs="name")
        if "defaultMetadataProfileId" in self._data:
            self.defaultMetadataProfileId = self._parse(attrs="defaultMetadataProfileId", value_type="int")
        if "defaultQualityProfileId" in self._data:
            self.defaultQualityProfileId = self._parse(attrs="defaultQualityProfileId", value_type="int")
        if "defaultMonitorOption" in self._data:
            self.defaultMonitorOption = self._parse(attrs="defaultMonitorOption", value_type="str")
        if "defaultTags" in self._data:
            self.defaultTags = self._parse(attrs="defaultTags", value_type="int", is_list=True)
        if "isCalibreLibrary" in self._data:
            self.isCalibreLibrary = self._parse(attrs="isCalibreLibrary", value_type="bool")
        self.freeSpace = self._parse(attrs="freeSpace", value_type="int")
        if "unmappedFolders" in self._data:
            self.unmappedFolders = [UnmappedFolder(self._arr, folder) for folder in self._data["unmappedFolders"]]
        self._finish(self.name if "name" in self._data else self.path)

    def delete(self):
        self._raw.delete_rootFolder(self.id)

class Season(SimpleObj):
    """ Represents a single Season.

        Attributes:
            seasonNumber (int): Season Number of the Tag.
            monitored (bool): If the Season is monitored.
            totalEpisodeCount (int): Total Episode Count for the Season.
            episodeCount (int): Episode Count for the Season.
            episodeFileCount (int): Episode File Count for the Season.
            sizeOnDisk (int): Size on Disk for the Season.
            percentOfEpisodes (float): Percent of Episodes for the Season..
            nextAiring (datetime): Next Airing Date for an Episode of this Season.
            previousAiring (datetime): Previous Airing Date for the latest Episode of this Season.
    """

    def _load(self, data):
        super()._load(data)
        self.seasonNumber = self._parse(attrs="seasonNumber", value_type="int")
        self.monitored = self._parse(attrs="monitored", value_type="bool")
        if "statistics" in self._data:
            self.totalEpisodeCount = self._parse(attrs=["statistics", "totalEpisodeCount"], value_type="int")
            self.episodeCount = self._parse(attrs=["statistics", "episodeCount"], value_type="int")
            self.episodeFileCount = self._parse(attrs=["statistics", "episodeFileCount"], value_type="int")
            self.sizeOnDisk = self._parse(attrs=["statistics", "sizeOnDisk"], value_type="int")
            self.percentOfEpisodes = self._parse(attrs=["statistics", "percentOfEpisodes"], value_type="float")
            self.nextAiring = self._parse(attrs="nextAiring", value_type="date")
            self.previousAiring = self._parse(attrs="previousAiring", value_type="date")
        self._finish(f"Season {self.seasonNumber}")


class UnmappedFolder(SimpleObj):
    """ Represents a single Unmapped Folder.

        Attributes:
            name (str): Name of the Unmapped Folder.
            path (str): Path of the Unmapped Folder.
    """

    def _load(self, data):
        super()._load(data)
        self.name = self._parse(attrs="name")
        self.path = self._parse(attrs="path")
        self._finish(self.path)

class RadarrExclusion(SimpleObj):
    """ Represents a single Radarr Exclusion.

        Attributes:
            tmdbId (int): TMDb ID of the Excluded Movie.
            title (str): Title of the Excluded Movie.
            year (int): Year of the Excluded Movie.
    """

    def _load(self, data):
        super()._load(data)
        self.tmdbId = self._parse(attrs="tmdbId", value_type="int")
        self.title = self._parse(attrs="movieTitle")
        self.year = self._parse(attrs="movieYear", value_type="int")
        self._finish(f"TMDb ID: {self.tmdbId}")

class SonarrExclusion(SimpleObj):
    """ Represents a single Sonarr Exclusion.

        Attributes:
            tvdbId (int): TVDb ID of the Excluded Series.
            title (str): Title of the Excluded Series.
    """

    def _load(self, data):
        super()._load(data)
        self.tvdbId = self._parse(attrs="tvdbId", value_type="int")
        self.title = self._parse(attrs="title")
        self._finish(f"TVDb ID: {self.tvdbId}")



from requests import Session
from typing import Optional
from .base import BaseRawAPI


class RadarrRawAPI(BaseRawAPI):
    def __init__(self, url: str, apikey: str, session: Optional[Session] = None) -> None:
        super().__init__(url, apikey, session=session)

    def get_movie(self, tmdb_id=None):
        """ GET /movie """
        if tmdb_id is not None:
            return self._get("movie", **{"tmdbId": tmdb_id})
        else:
            return self._get("movie")

    def get_movie_id(self, movie_id):
        """ GET /movie/{id} """
        return self._get(f"movie/{movie_id}")

    def post_movie(self, json):
        """ POST /movie """
        return self._post("movie", json=json)

    def post_movie_import(self, json):
        """ POST /movie/import """
        return self._post("movie/import", json=json)

    def put_movie(self, json, moveFiles=False):
        """ PUT /movie """
        params = {"moveFiles": "true"} if moveFiles else {}
        return self._put("movie", json=json, **params)

    def put_movie_id(self, movie_id, json, moveFiles=False):
        """ PUT /movie/{id} """
        params = {"moveFiles": "true"} if moveFiles else {}
        return self._put(f"movie/{movie_id}", json=json, **params)

    def put_movie_editor(self, json):
        """ PUT /movie/editor """
        return self._put("movie/editor", json=json)

    def delete_movie_id(self, movie_id, addImportExclusion=False, deleteFiles=False):
        """ DELETE /movie/{id} """
        params = {}
        if addImportExclusion:
            params["addImportExclusion"] = "true"
        if deleteFiles:
            params["deleteFiles"] = "true"
        self._delete(f"movie/{movie_id}", **params)

    def delete_movie_editor(self, json):
        """ DELETE /movie/editor """
        return self._delete("movie/editor", json=json)

    def get_movie_lookup(self, term):
        """ GET /movie/lookup """
        return self._get("movie/lookup", **{"term": term})

    def get_exclusions(self):
        """ GET /exclusions """
        return self._get("exclusions")

    def post_exclusions(self, json):
        """ POST /exclusions """
        return self._post("exclusions", json=json)

    def add_exclusion(self, title, tmdb_id, year):
        return self.post_exclusions({
            "movieTitle": title,
            "tmdbId": tmdb_id,
            "movieYear": year
        })



from requests import Session
from typing import Optional, Union, List, Tuple
from arrapi import LanguageProfile, RootFolder, QualityProfile, Series, Tag, NotFound, Invalid, Exists
from .base import BaseAPI
from ..exceptions import Excluded
from ..objs.simple import SonarrExclusion
from ..raws.sonarr import SonarrRawAPI


class SonarrAPI(BaseAPI):
    """ Primary Class to use when connecting with the Sonarr API

        Parameters:
            url (str): URL of Sonarr application.
            apikey (str): apikey for the Sonarr application.
            session (Optional[Session]): Session object to use.
     """

    def __init__(self, url: str, apikey: str, session: Optional[Session] = None) -> None:
        super().__init__(SonarrRawAPI(url, apikey, session=session))
        self.exclusions = []
        self.monitor_options = ["all", "future", "missing", "existing", "pilot", "firstSeason", "latestSeason", "none"]
        self.series_type_options = ["standard", "daily", "anime"]

    def _validate_add_options(self, root_folder, quality_profile, language_profile=None, monitor="all",
                              season_folder=True, search=True, unmet_search=False, series_type="standard",
                              tags=None):
        """ Validate Add Series options. """
        options = {
            "root_folder": self._validate_root_folder(root_folder),
            "quality_profile" if self._raw.new_codebase else "profileId": self._validate_quality_profile(quality_profile),
            "monitor": self._validate_monitor(monitor),
            "monitored": monitor != "none",
            "season_folder": True if season_folder is True else False,
            "search": True if search is True else False,
            "unmet_search": True if unmet_search is True else False,
            "series_type": self._validate_series_type(series_type),
        }
        if not self._raw.v4:
            if not language_profile:
                raise Invalid("Language Profile Required")
            options["language_profile"] = self._validate_language_profile(language_profile)
        if tags:
            options["tags"] = self._validate_tags(tags)
        return options

    def _validate_edit_options(self, root_folder=None, path=None, move_files=False, quality_profile=None,
                               language_profile=None, monitor=None, monitored=None, season_folder=None,
                               series_type=None, tags=None, apply_tags="add"):
        """ Validate Edit Series options. """
        variables = [root_folder, path, quality_profile, language_profile, monitor,
                     monitored, season_folder, series_type, tags]
        if all(v is None for v in variables):
            raise ValueError("Expected either root_folder, path, quality_profile, language_profile, "
                             "monitor, monitored, season_folder, series_type, or tags args")
        options = {"moveFiles": True if move_files is True else False}
        if root_folder is not None:
            options["rootFolderPath"] = self._validate_root_folder(root_folder)
        if path is not None:
            options["path"] = path
        if quality_profile is not None:
            options["qualityProfileId" if self._raw.new_codebase else "profileId"] = self._validate_quality_profile(quality_profile)
        if language_profile is not None and not self._raw.v4:
            options["languageProfileId"] = self._validate_language_profile(language_profile)
        if monitor is not None:
            options["monitor"] = self._validate_monitor(monitor)
        if monitored is not None:
            options["monitored"] = True if monitored is True else False
        if season_folder is not None:
            options["seasonFolder"] = True if season_folder is True else False
        if series_type is not None:
            options["seriesType"] = self._validate_series_type(series_type)
        if tags is not None:
            options["tags"] = self._validate_tags(tags, create=apply_tags != "remove")
            if apply_tags in self.apply_tags_options:
                options["applyTags"] = apply_tags
            else:
                raise Invalid(f"Invalid apply_tags: '{apply_tags}' Options: {self.apply_tags_options}")
        return options

    def _validate_monitor(self, monitor):
        """ Validate Monitor options. """
        return self._validate_options("Monitor", monitor, self.monitor_options)

    def _validate_series_type(self, series_type):
        """ Validate Series Type options. """
        return self._validate_options("Series Type", series_type, self.series_type_options)

    def _validate_tvdb_ids(self, ids):
        """ Validate TVDb IDs. """
        valid_ids = []
        invalid_ids = []
        used_ids = []
        sonarr_ids = {}
        for s in self.all_series():
            sonarr_ids[s.tvdbId] = s
            sonarr_ids[str(s.tvdbId)] = s
        for _id in ids:
            if isinstance(_id, Series) and str(_id.tvdbId) not in used_ids:
                valid_ids.append(_id.id)
                used_ids.append(str(_id.tvdbId))
            elif _id in sonarr_ids and str(_id) not in used_ids:
                valid_ids.append(sonarr_ids[_id].id)
                used_ids.append(str(_id))
            else:
                invalid_ids.append(_id)
        return valid_ids, invalid_ids

    def respect_list_exclusions_when_adding(self):
        """ Stores all List Exclusions so whenever :func:`~arrapi.objs.reload.Series.add` or :func:`~arrapi.apis.sonarr.SonarrAPI.add_multiple_series` is called the additions will be checked against the Exclusion List  """
        self.exclusions = [SonarrExclusion(self, ex).tvdbId for ex in self._raw.get_importlistexclusion()]

    def get_series(self, series_id: Optional[int] = None, tvdb_id: Optional[int] = None) -> Series:
        """ Gets a :class:`~arrapi.objs.reload.Series` by one of the IDs.

            Parameters:
                series_id (Optional[int]): Search by Sonarr Series ID.
                tvdb_id (Optional[int]): Search by TVDb ID.

            Returns:
                :class:`~arrapi.objs.reload.Series`: Series for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given.
                :class:`~arrapi.exceptions.NotFound`: When there's no series with that ID.
        """
        if all(v is None for v in [series_id, tvdb_id]):
            raise ValueError("Expected either series_id or tvdb_id args")
        return Series(self, series_id=series_id, tvdb_id=tvdb_id)

    def all_series(self) -> List[Series]:
        """ Gets all :class:`~arrapi.objs.reload.Series` in Sonarr.

            Returns:
                List[:class:`~arrapi.objs.reload.Series`]: List of Series in Sonarr.
        """
        return [Series(self, data=d) for d in self._raw.get_series()]

    def search_series(self, term: str) -> List[Series]:
        """ Gets a list of :class:`~arrapi.objs.reload.Series` by a search term.

            Parameters:
                term (str): Term to Search for.

            Returns:
                List[:class:`~arrapi.objs.reload.Series`]: List of Series's found.
        """
        return [Series(self, data=d) for d in self._raw.get_series_lookup(term)]

    def add_series(
            self,
            root_folder: Union[str, int, "RootFolder"],
            quality_profile: Union[str, int, "QualityProfile"],
            language_profile: Union[str, int, "LanguageProfile"],
            series_id: Optional[int] = None,
            tvdb_id: Optional[int] = None,
            monitor: str = "all",
            season_folder: bool = True,
            search: bool = True,
            unmet_search: bool = True,
            series_type: str = "standard",
            tags: Optional[List[Union[str, int, Tag]]] = None
    ) -> Series:
        """ Gets a :class:`~arrapi.objs.reload.Series` by one of the IDs and adds it to Sonarr.

            Parameters:
                root_folder (Union[str, int, RootFolder]): Root Folder for the Series.
                quality_profile (Union[str, int, QualityProfile]): Quality Profile for the Series.
                language_profile (Union[str, int, LanguageProfile]): Language Profile for the Series.
                series_id (Optional[int]): Search by Sonarr Series ID.
                tvdb_id (Optional[int]): Search by TVDb ID.
                monitor (bool): How to monitor the Series. Valid options are all, future, missing, existing, pilot, firstSeason, latestSeason, or none.
                season_folder (bool): Use Season Folders for the Series.
                search (bool): Start search for missing episodes of the Series after adding.
                unmet_search (bool): Start search for cutoff unmet episodes of the Series after adding.
                series_type (str): Series Type for the Series. Valid options are standard, daily, or anime.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added to the Series.

            Returns:
                :class:`~arrapi.objs.reload.Series`: Series for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given.
                :class:`~arrapi.exceptions.NotFound`: When there's no series with that ID.
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.Exists`: When the Series already Exists in Sonarr.
                :class:`~arrapi.exceptions.Excluded`: When the Series Exists in the Exclusion List.
        """
        series = self.get_series(series_id=series_id, tvdb_id=tvdb_id)
        series.add(root_folder, quality_profile, language_profile, monitor=monitor, season_folder=season_folder,
                   search=search, unmet_search=unmet_search, series_type=series_type, tags=tags)
        return series

    def edit_series(
            self,
            series_id: Optional[int] = None,
            tvdb_id: Optional[int] = None,
            path: Optional[str] = None,
            move_files: bool = False,
            quality_profile: Optional[Union[str, int, "QualityProfile"]] = None,
            language_profile: Optional[Union[str, int, "LanguageProfile"]] = None,
            monitor: Optional[str] = None,
            monitored: Optional[bool] = None,
            season_folder: Optional[bool] = None,
            series_type: Optional[str] = None,
            tags: Optional[List[Union[str, int, Tag]]] = None,
            apply_tags: str = "add"
            ) -> Series:
        """ Gets a :class:`~arrapi.objs.reload.Series` by one of the IDs and edits it in Sonarr.

            Parameters:
                series_id (Optional[int]): Search by Sonarr Series ID.
                tvdb_id (Optional[int]): Search by TVDb ID.
                path (Optional[str]): Path to change the Series to.
                move_files (bool): When changing the path do you want to move the files to the new path.
                quality_profile (Optional[Union[str, int, QualityProfile]]): Quality Profile to change the Series to.
                language_profile (Optional[Union[str, int, LanguageProfile]]): Language Profile to change the Series to.
                monitor (Optional[str]): How you want the Series monitored. Valid options are all, future, missing, existing, pilot, firstSeason, latestSeason, or none.
                monitored (Optional[bool]): Monitor the Series.
                season_folder (Optional[bool]): Use Season Folders for the Series.
                series_type (Optional[str]): Series Type to change the Series to. Valid options are standard, daily, or anime.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added, replaced, or removed from the Series.
                apply_tags (str): How you want to edit the Tags. Valid options are add, replace, or remove.

            Returns:
                :class:`~arrapi.objs.reload.Series`: Series for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given or when there's no options given.
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
                :class:`~arrapi.exceptions.NotFound`: When there's no series with that ID or when the Series hasn't been added to Sonarr.
        """
        series = self.get_series(series_id=series_id, tvdb_id=tvdb_id)
        series.edit(path=path, move_files=move_files, quality_profile=quality_profile,
                    language_profile=language_profile, monitor=monitor, monitored=monitored,
                    season_folder=season_folder, series_type=series_type, tags=tags, apply_tags=apply_tags)
        return series

    def delete_series(
            self,
            series_id: Optional[int] = None,
            tvdb_id: Optional[int] = None,
            addImportListExclusion: bool = False,
            deleteFiles: bool = False
    ) -> Series:
        """ Gets a :class:`~arrapi.objs.reload.Series` by one of the IDs and deletes it from Sonarr.

            Parameters:
                series_id (Optional[int]): Search by Sonarr Series ID.
                tvdb_id (Optional[int]): Search by TVDb ID.
                addImportListExclusion (bool): Add Import Exclusion for this Series.
                deleteFiles (bool): Delete Files for this Series.

            Returns:
                :class:`~arrapi.objs.reload.Series`: Series for the ID given.

            Raises:
                :class:`ValueError`: When no ID is given.
                :class:`~arrapi.exceptions.NotFound`: When there's no series with that ID or when the Series hasn't been added to Sonarr.
        """
        series = self.get_series(series_id=series_id, tvdb_id=tvdb_id)
        series.delete(addImportListExclusion=addImportListExclusion, deleteFiles=deleteFiles)
        return series

    def add_multiple_series(self, ids: List[Union[Series, int, Tuple[Union[Series, int], str]]],
                            root_folder: Union[str, int, RootFolder],
                            quality_profile: Union[str, int, QualityProfile],
                            language_profile: Optional[Union[str, int, LanguageProfile]] = None,
                            monitor: str = "all",
                            season_folder: bool = True,
                            search: bool = True,
                            unmet_search: bool = True,
                            series_type: str = "standard",
                            tags: Optional[List[Union[str, int, Tag]]] = None,
                            per_request: int = None
                            ) -> Tuple[List[Series], List[Series], List[Union[int, Series]], List[int]]:
        """ Adds multiple Series to Sonarr in a single call by their TVDb IDs.

            You can specify the path for each TVDb ID using a tuple in the list instead of just the ID ex. ``(121361, "/media/Game of Thrones/")``

            The path provided must begin with the root_folder specified.

            Parameters:
                ids (List[Union[Series, int, Tuple[Union[Series, int], str]]]): List of TVDB IDs or Series lookups to add.
                root_folder (Union[str, int, RootFolder]): Root Folder for the Series.
                quality_profile (Union[str, int, QualityProfile]): Quality Profile for the Series.
                language_profile (Optional[Union[str, int, LanguageProfile]]): Language Profile for the Series. Required for older versions only.
                monitor (bool): How to monitor the Series. Valid options are ``all``, ``future``, ``missing``, ``existing``, ``pilot``, ``firstSeason``, ``latestSeason``, or ``none``.
                season_folder (bool): Use Season Folders for the Series.
                search (bool): Start search for missing episodes of the Series after adding.
                unmet_search (bool): Start search for cutoff unmet episodes of the Series after adding.
                series_type (str): Series Type for the Series. Valid options are ``standard``, ``daily``, or ``anime``.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added to the Series.
                per_request (int): Number of Series to add per request.

            Returns:
                Tuple[List[:class:`~arrapi.objs.reload.Series`], List[:class:`~arrapi.objs.reload.Series`], List[Union[int, Series]], List[int]]: List of Series that were able to be added, List of Series already in Sonarr, List of Series that could not be found, List of Movies that were excluded.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
        """
        options = self._validate_add_options(root_folder, quality_profile, language_profile=language_profile,
                                             monitor=monitor, season_folder=season_folder, search=search,
                                             unmet_search=unmet_search, series_type=series_type, tags=tags)
        json = []
        series = []
        existing_series = []
        invalid_ids = []
        excluded_ids = []
        used_ids = []
        for input_item in ids:
            path = input_item[1] if isinstance(input_item, tuple) else None
            item = input_item[0] if isinstance(input_item, tuple) else input_item
            try:
                if isinstance(item, Series):
                    show = item
                else:
                    if int(item) in used_ids or (self.exclusions and int(item) in self.exclusions):
                        raise Excluded(int(item))
                    show = self.get_series(tvdb_id=item)
                if show.tvdbId in used_ids or (self.exclusions and show.tvdbId in self.exclusions):
                    raise Excluded(show.tvdbId)
                used_ids.append(show.tvdbId)
                try:
                    json.append(show._get_add_data(options, path=path))
                except Exists:
                    existing_series.append(show)
            except NotFound:
                invalid_ids.append(input_item)
            except Excluded as e:
                excluded_ids.append(int(str(e)))
        if len(json) > 0:
            if per_request is None:
                per_request = len(json)
            for i in range(0, len(json), per_request):
                series.extend([Series(self, data=s) for s in self._raw.post_series_import(json[i:i+per_request])])
        return series, existing_series, invalid_ids, excluded_ids

    def edit_multiple_series(self, ids: List[Union[Series, int]],
                             root_folder: Optional[Union[str, int, RootFolder]] = None,
                             move_files: bool = False,
                             quality_profile: Optional[Union[str, int, QualityProfile]] = None,
                             language_profile: Optional[Union[str, int, LanguageProfile]] = None,
                             monitor: Optional[str] = None,
                             monitored: Optional[bool] = None,
                             season_folder: Optional[bool] = None,
                             series_type: Optional[str] = None,
                             tags: Optional[List[Union[str, int, Tag]]] = None,
                             apply_tags: str = "add",
                             per_request: int = None
                             ) -> Tuple[List[Series], List[Union[Series, int]]]:
        """ Edit multiple Series in Sonarr by their TVDb IDs.

            Parameters:
                ids (List[Union[int, Series]]): List of Series IDs or Series objects you want to edit.
                root_folder (Union[str, int, RootFolder]): Root Folder to change the Series to.
                move_files (bool): When changing the root folder do you want to move the files to the new path.
                quality_profile (Optional[Union[str, int, QualityProfile]]): Quality Profile to change the Series to.
                language_profile (Optional[Union[str, int, LanguageProfile]]): Language Profile to change the Series to.
                monitor (Optional[str]): How you want the Series monitored. Valid options are all, future, missing, existing, pilot, firstSeason, latestSeason, or none.
                monitored (Optional[bool]): Monitor the Series.
                season_folder (Optional[bool]): Use Season Folders for the Series.
                series_type (Optional[str]): Series Type to change the Series to. Valid options are standard, daily, or anime.
                tags (Optional[List[Union[str, int, Tag]]]): Tags to be added, replaced, or removed from the Series.
                apply_tags (str): How you want to edit the Tags. Valid options are add, replace, or remove.
                per_request (int): Number of Series to edit per request.

            Returns:
                Tuple[List[:class:`~arrapi.objs.reload.Series`], List[Union[Series, int]]]: List of Series that were able to be edited, List of Series that could not be found in Sonarr.

            Raises:
                :class:`~arrapi.exceptions.Invalid`: When one of the options given is invalid.
        """
        json = self._validate_edit_options(root_folder=root_folder, move_files=move_files,
                                           quality_profile=quality_profile, language_profile=language_profile,
                                           monitor=monitor, monitored=monitored, season_folder=season_folder,
                                           series_type=series_type, tags=tags, apply_tags=apply_tags)
        series_list = []
        valid_ids, invalid_ids = self._validate_tvdb_ids(ids)
        if len(valid_ids) > 0:
            if per_request is None:
                per_request = len(valid_ids)
            if "monitor" in json:
                json_monitor = json.pop("monitor")
                for i in range(0, len(valid_ids), per_request):
                    self._raw.edit_series_monitoring(valid_ids[i:i+per_request], json_monitor)
            for i in range(0, len(valid_ids), per_request):
                json["seriesIds"] = valid_ids[i:i+per_request]
                series_list.extend([Series(self, data=s) for s in self._raw.put_series_editor(json)])
        return series_list, invalid_ids

    def delete_multiple_series(self, ids: List[Union[int, Series]],
                               addImportExclusion: bool = False,
                               deleteFiles: bool = False,
                               per_request: int = None
                               ) -> List[Union[Series, int]]:
        """ Deletes multiple Series in Sonarr by their TVDb IDs.

            Parameters:
                ids (List[Union[int, Series]]): List of TVDb IDs or Series objects you want to delete.
                addImportExclusion (bool): Add Import Exclusion for these TVDb IDs.
                deleteFiles (bool): Delete Files for these TVDb IDs.
                per_request (int): Number of Series to delete per request.

            Returns:
                List[Union[Series, int]]: List of Series that could not be found in Sonarr.
        """
        valid_ids, invalid_ids = self._validate_tvdb_ids(ids)
        if len(valid_ids) > 0:
            json = {
                "deleteFiles": deleteFiles,
                "addImportExclusion": addImportExclusion
            }
            if per_request is None:
                per_request = len(valid_ids)
            for i in range(0, len(valid_ids), per_request):
                json["seriesIds"] = valid_ids[i:i+per_request]
                self._raw.delete_series_editor(json)
        return invalid_ids

    def language_profile(self) -> List[LanguageProfile]:
        """ Gets every :class:`~arrapi.objs.reload.LanguageProfile` in Sonarr.

            Returns:
                List[:class:`~arrapi.objs.reload.LanguageProfile`]: List of all Language Profiles
        """
        return [LanguageProfile(self, data) for data in self._raw.get_languageProfile()]

    def _validate_language_profile(self, language_profile):
        """ Validate Quality Profile options. """
        options = []
        for profile in self.language_profile():
            options.append(profile)
            if (isinstance(language_profile, LanguageProfile) and profile.id == language_profile.id) \
                    or (isinstance(language_profile, int) and profile.id == language_profile) \
                    or (profile.name == language_profile):
                return profile.id
        raise Invalid(f"Invalid Language Profile: '{language_profile}' Options: {options}")




import os, unittest

from arrapi import SonarrAPI, NotFound, Exists
"""
import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())
"""
class SonarrTests(unittest.TestCase):
    sonarr = None
    root = "/config"
    profile = "HD-1080p"
    profile2 = "HD - 720p/1080p"
    language = "English"

    @classmethod
    def setUpClass(cls):
        cls.sonarr = SonarrAPI(os.environ["SONARR_URL"], os.environ["SONARR_APIKEY"])
        has_config = False
        for rf in cls.sonarr.root_folder():
            if rf.path == cls.root:
                has_config = True
        if not has_config:
            cls.sonarr.add_root_folder(f"{cls.root}/")
        for series in cls.sonarr.all_series():
            series.delete()
        for tag in cls.sonarr.all_tags():
            tag.delete()

    def test_lookups(self):
        series = self.sonarr.get_series(tvdb_id=121361)
        self.assertEqual(series.title, "Game of Thrones")
        self.assertEqual(series.tvdbId, 121361)
        self.assertEqual(series.imdbId, "tt0944947")
        search = self.sonarr.search_series("Breaking Bad")
        self.assertEqual(search[0].title, "Breaking Bad")

    def get_test_series(self):
        return [series for series in self.sonarr.all_series() if series.tvdbId in [83268, 283468, 385376]]

    def test_multiple_add_edit_delete(self):
        series_ids = [self.sonarr.get_series(tvdb_id=83268), 283468, 385376]
        self.sonarr.add_multiple_series(series_ids, self.root, self.profile, self.language, tags=["firsttag"])
        test_series = self.get_test_series()
        tvdb_ids = [series.tvdbId for series in test_series]
        self.assertIn(83268, tvdb_ids)
        self.assertIn(283468, tvdb_ids)
        self.assertIn(385376, tvdb_ids)
        for series in test_series:
            self.assertEqual(len(series.tags), 1)
            self.assertEqual(series.tags[0].label, "firsttag")
        series_ids[0].reload()
        self.sonarr.edit_multiple_series(series_ids, tags=["secondtag"], apply_tags="replace")
        for series in self.get_test_series():
            self.assertEqual(len(series.tags), 1)
            self.assertEqual(series.tags[0].label, "secondtag")
        self.sonarr.delete_multiple_series(series_ids)
        self.assertEqual(len(self.get_test_series()), 0)

    def test_single_add_edit_delete(self):
        series = self.sonarr.get_series(tvdb_id=121361)
        with self.assertRaises(NotFound):
            series.delete()
        self.assertEqual(series.title, "Game of Thrones")
        self.assertEqual(series.tvdbId, 121361)
        self.assertIsNone(series.id)
        series.add(self.root, self.profile, self.language, tags=["arrapi"])
        with self.assertRaises(Exists):
            series.add(self.root, self.profile, self.language)
        self.assertIsNotNone(series.id)
        series = self.sonarr.get_series(series_id=series.id)
        self.assertGreater(len(series.tags), 0)
        self.assertEqual(series.tags[0].label, "arrapi")
        self.assertEqual(series.seriesType, "standard")
        self.assertEqual(series.qualityProfile.name, self.profile)
        series.edit(series_type="anime", quality_profile=self.profile2, tags=["arrapi"], apply_tags="remove")
        self.assertEqual(len(series.tags), 0)
        self.assertEqual(series.seriesType, "anime")
        self.assertEqual(series.qualityProfile.name, self.profile2)
        series.delete()
        self.assertIsNone(series.id)
        with self.assertRaises(NotFound):
            series.delete()

    def test_tag_edit(self):
        tag = next((t for t in self.sonarr.all_tags() if t.label == "arrapi"), None)
        self.assertIsNotNone(tag)
        tag.edit("newtag")
        tag = next((t for t in self.sonarr.all_tags() if t.label == "newtag"), None)
        self.assertIsNotNone(tag)








import os, unittest

from arrapi import RadarrAPI, NotFound, Exists
"""
import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())
"""
class RadarrTests(unittest.TestCase):
    radarr = None
    root = "/config"
    profile = "HD-1080p"
    profile2 = "HD - 720p/1080p"

    @classmethod
    def setUpClass(cls):
        cls.radarr = RadarrAPI(os.environ["RADARR_URL"], os.environ["RADARR_APIKEY"])
        has_config = False
        for rf in cls.radarr.root_folder():
            if rf.path == cls.root:
                has_config = True
        if not has_config:
            cls.radarr.add_root_folder(f"{cls.root}/")
        for movie in cls.radarr.all_movies():
            movie.delete()
        for tag in cls.radarr.all_tags():
            tag.delete()

    def test_lookups(self):
        movie = self.radarr.get_movie(imdb_id="tt0080684")
        self.assertEqual(movie.title, "The Empire Strikes Back")
        self.assertEqual(movie.tmdbId, 1891)
        self.assertEqual(movie.imdbId, "tt0080684")
        search = self.radarr.search_movies("Return of the Jedi")
        self.assertEqual(search[0].title, "Return of the Jedi")

    def get_test_movies(self):
        return [movie for movie in self.radarr.all_movies() if movie.tmdbId in [11, 1891, 1892, 1893, 1894, 1895]]

    def test_multiple_add_edit_delete(self):
        movie_ids = [self.radarr.get_movie(tmdb_id=11), "tt0080684", 1892, 1893, 1894, 1895]
        self.radarr.add_multiple_movies(movie_ids, self.root, self.profile, tags=["firsttag"])
        test_movies = self.get_test_movies()
        tmdb_ids = [movie.tmdbId for movie in test_movies]
        self.assertIn(11, tmdb_ids)
        self.assertIn(1891, tmdb_ids)
        self.assertIn(1892, tmdb_ids)
        self.assertIn(1893, tmdb_ids)
        self.assertIn(1894, tmdb_ids)
        self.assertIn(1895, tmdb_ids)
        for movie in test_movies:
            self.assertEqual(len(movie.tags), 1)
            self.assertEqual(movie.tags[0].label, "firsttag")
        movie_ids[0].reload()
        self.radarr.edit_multiple_movies(movie_ids, tags=["secondtag"], apply_tags="replace")
        for movie in self.get_test_movies():
            self.assertEqual(len(movie.tags), 1)
            self.assertEqual(movie.tags[0].label, "secondtag")
        self.radarr.delete_multiple_movies(movie_ids)
        self.assertEqual(len(self.get_test_movies()), 0)

    def test_single_add_edit_delete(self):
        movie = self.radarr.get_movie(tmdb_id=11)
        with self.assertRaises(NotFound):
            movie.delete()
        self.assertEqual(movie.title, "Star Wars")
        self.assertEqual(movie.tmdbId, 11)
        self.assertIsNone(movie.id)
        movie.add(self.root, self.profile, tags=["arrapi"])
        with self.assertRaises(Exists):
            movie.add(self.root, self.profile)
        self.assertIsNotNone(movie.id)
        movie = self.radarr.get_movie(movie_id=movie.id)
        self.assertGreater(len(movie.tags), 0)
        self.assertEqual(movie.tags[0].label, "arrapi")
        self.assertEqual(movie.minimumAvailability, "announced")
        self.assertEqual(movie.qualityProfile.name, self.profile)
        movie.edit(minimum_availability="released", quality_profile=self.profile2, tags=["arrapi"], apply_tags="remove")
        self.assertEqual(len(movie.tags), 0)
        self.assertEqual(movie.minimumAvailability, "released")
        self.assertEqual(movie.qualityProfile.name, self.profile2)
        movie.delete()
        self.assertIsNone(movie.id)
        with self.assertRaises(NotFound):
            movie.delete()

    def test_tag_edit(self):
        tag = next((t for t in self.radarr.all_tags() if t.label == "arrapi"), None)
        self.assertIsNotNone(tag)
        tag.edit("newtag")
        tag = next((t for t in self.radarr.all_tags() if t.label == "newtag"), None)
        self.assertIsNotNone(tag)








from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, Union, Any


class BaseObj(ABC):
    """ Base Class for Arr Objects.

        Attributes:
            id (int): ID of the Object.
    """

    def __init__(self, arr, data):
        self._loading = True
        self._arr = arr
        self._raw = arr._raw
        self._partial = False
        self._name = None
        self._load(data)

    @abstractmethod
    def _load(self, data):
        self._data = data
        self._loading = True
        self.id = None

    def _finish(self, name):
        self._name = name
        self._loading = False

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"[{self.id}:{self._name}]" if self.id is not None else f"[{self._name}]"

    def __eq__(self, other):
        if type(self) is type(other):
            if self.id is None and other.id is None:
                return self._name == other._name
            elif self.id is not None and other.id is not None:
                return self.id == other.id
            else:
                return False
        elif isinstance(other, int) and self.id is not None:
            return self.id == other
        else:
            return str(self._name) == str(other)

    def __getattribute__(self, item):
        value = super().__getattribute__(item)
        if item.startswith("_") or self._loading or not self._partial or value is not None:
            return value
        self._load(None)
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if key.startswith("_") or self._loading:
            super().__setattr__(key, value)
        else:
            raise AttributeError("Attributes cannot be edited")

    def __delattr__(self, key):
        raise AttributeError("Attributes cannot be deleted")

    def _parse(self, data: Any = None, attrs: Optional[Union[str, list]] = None, value_type: str = "str",
               default_is_none: bool = False, is_list: bool = False):
        """ Validate the value given from the options given.

            Parameters:
                data (Any): data to check
                attrs (Optional[Union[str, list]]): check data for these attributes.
                value_type (str): Type that the value is.
                default_is_none (bool): Makes default None.
                is_list (bool): Is list of values

            Returns:
                Any: Parsed Value
        """
        if default_is_none is False and value_type in ["int", "float"]:
            default = 0
        elif default_is_none is False and value_type.endswith("List"):
            default = []
        else:
            default = None

        value = self._data if data is None else data
        if attrs:
            if not isinstance(attrs, list):
                attrs = [attrs]
            for attr in attrs:
                if attr in value:
                    value = value[attr]
                else:
                    return default

        if value is None:
            return default
        elif is_list:
            return [self._parse(data=v, value_type=value_type, default_is_none=default_is_none) for v in value]
        elif value_type == "int":
            return int(value)
        elif value_type == "float":
            return float(value)
        elif value_type == "bool":
            if isinstance(value, bool):
                return value
            elif str(value).lower() in ["t", "true"]:
                return True
            elif str(value).lower() in ["f", "false"]:
                return False
            else:
                return default
        elif value_type == "date":
            return datetime.strptime(value[:-1].split(".")[0], "%Y-%m-%dT%H:%M:%S")
        elif value_type == "collection":
            return arrapi.objs.simple.Collection(self._arr, value)
        elif value_type == "image":
            return arrapi.objs.simple.Image(self._arr, value)
        elif value_type == "season":
            return arrapi.objs.simple.Season(self._arr, value)
        elif value_type == "intTag":
            return arrapi.objs.reload.Tag(self._arr, {"id": value})
        elif value_type == "intQualityProfile":
            return arrapi.objs.reload.QualityProfile(self._arr, {"id": value})
        elif value_type == "intLanguageProfile":
            return arrapi.objs.reload.LanguageProfile(self._arr, {"id": value})
        elif value_type == "str":
            return str(value)
        else:
            return value

import arrapi.objs.simple
import arrapi.objs.reload
